import kernpy as kp

import unittest


class MxhmSpineImporterTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.doc_with_mxhm, _ = kp.load('resource_dir/samples/jazzmus_with_mxhm.krn')

    def do_test_token_exported(self, input_encoding, expected):
        importer = kp.MxhmSpineImporter()
        token = importer.import_token(input_encoding)
        self.assertIsNotNone(token)
        self.assertEqual(expected, token.export())

        return token

    def do_test_token_category(self, input_encoding, expected_category):
        importer = kp.MxhmSpineImporter()
        token = importer.import_token(input_encoding)
        self.assertIsNotNone(token)
        self.assertEqual(expected_category, token.category)

        return token





from __future__ import annotations

import csv
import io
from copy import copy
from pathlib import Path
from typing import List, Optional

from kernpy.core.tokens import TokenCategory, SignatureToken, MetacommentToken, HeaderToken, SpineOperationToken, \
    FieldCommentToken, ErrorToken, \
    BoundingBoxToken, SPINE_OPERATIONS, HEADERS, Token
from kernpy.core.document import Document, MultistageTree, BoundingBoxMeasures
from kernpy.core.importer_factory import createImporter


class Importer:
    """
    Importer class.

    Use this class to import the content from a file or a string to a `Document` object.
    """
    def __init__(self):
        """
        Create an instance of the importer.

        Raises:
            Exception: If the importer content is not a valid **kern file.

        Examples:
            # Create the importer
            >>> importer = Importer()

            # Import the content from a file
            >>> document = importer.import_file('file.krn')

            # Import the content from a string
            >>> document = importer.import_string("**kern\n*clefF4\nc4\n4d\n4e\n4f\n*-")
        """
        self.last_measure_number = None
        self.last_bounding_box = None
        self.errors = []

        self._tree = MultistageTree()
        self._document = Document(self._tree)
        self._importers = {}
        self._header_row_number = None
        self._row_number = 1
        self._tree_stage = 0
        self._next_stage_parents = None
        self._prev_stage_parents = None
        self._last_node_previous_to_header = self._tree.root

    @staticmethod
    def get_last_spine_operator(parent):
        if parent is None:
            return None
        elif isinstance(parent.token, SpineOperationToken):
            return parent
        else:
            return parent.last_spine_operator_node

    #TODO Documentar cómo propagamos los header_node y last_spine_operator_node...
    def run(self, reader) -> Document:
        for row in reader:
            if len(row) <= 0:
                # Found an empty row, usually the last one. Ignore it.
                continue

            self._tree_stage = self._tree_stage + 1
            is_barline = False
            if self._next_stage_parents:
                self._prev_stage_parents = copy(self._next_stage_parents)
            self._next_stage_parents = []

            if row[0].startswith("!!"):
                self._compute_metacomment_token(row[0].strip())
            else:
                for icolumn, column in enumerate(row):
                    if column.startswith("**"):
                        self._compute_header_token(icolumn, column)
                        # go to next row
                        continue

                    if column in SPINE_OPERATIONS:
                        self._compute_spine_operator_token(icolumn, column, row)
                    else:  # column is not a spine operation
                        if column.startswith("!"):
                            token = FieldCommentToken(column)
                        else:
                            if self._prev_stage_parents is None:
                                raise ValueError(f'Any spine header found in the column #{icolumn}. '
                                                 f'Expected a previous line with valid content. '
                                                 f'The token in column #{icolumn} and row #{self._row_number - 1}'
                                                 f' was not created correctly. Error detected in '
                                                 f'column #{icolumn} in row #{self._row_number}. '
                                                 f'Found {column}. ')
                            if icolumn >= len(self._prev_stage_parents):
                                # TODO: Try to fix the kern in runtime. Add options to public API
                                # continue  # ignore the column
                                raise ValueError(f'Wrong columns number in row {self._row_number}. '
                                                 f'The token in column #{icolumn} and row #{self._row_number}'
                                                 f' has more columns than expected in its row. '
                                                 f'Expected {len(self._prev_stage_parents)} columns '
                                                 f'but found {len(row)}.')
                            parent = self._prev_stage_parents[icolumn]
                            if not parent:
                                raise Exception(f'Cannot find a parent node for column #{icolumn} in row {self._row_number}')
                            if not parent.header_node:
                                raise Exception(f'Cannot find a header node for column #{icolumn} in row {self._row_number}')
                            importer = self._importers.get(parent.header_node.token.encoding)
                            if not importer:
                                raise Exception(f'Cannot find an importer for header {parent.header_node.token.encoding}')
                            try:
                                token = importer.import_token(column)
                            except Exception as error:
                                token = ErrorToken(column, self._row_number, str(error))
                                self.errors.append(token)
                        if not token:
                            raise Exception(
                                f'No token generated for input {column} in row number #{self._row_number} using importer {importer}')

                        parent = self._prev_stage_parents[icolumn]
                        node = self._tree.add_node(self._tree_stage, parent, token, self.get_last_spine_operator(parent), parent.last_signature_nodes, parent.header_node)
                        self._next_stage_parents.append(node)

                        if (token.category == TokenCategory.BARLINES
                                or TokenCategory.is_child(child=token.category, parent=TokenCategory.CORE)
                                    and len(self._document.measure_start_tree_stages) == 0):
                            is_barline = True
                        elif isinstance(token, BoundingBoxToken):
                            self.handle_bounding_box(self._document, token)
                        elif isinstance(token, SignatureToken):
                            node.last_signature_nodes.update(node)

                if is_barline:
                    self._document.measure_start_tree_stages.append(self._tree_stage)
                    self.last_measure_number = len(self._document.measure_start_tree_stages)
                    if self.last_bounding_box:
                        self.last_bounding_box.to_measure = self.last_measure_number
            self._row_number = self._row_number + 1
        return self._document

    def handle_bounding_box(self, document: Document, token: BoundingBoxToken):
        page_number = token.page_number
        last_page_bb = document.page_bounding_boxes.get(page_number)
        if last_page_bb is None:
            if self.last_measure_number is None:
                self.last_measure_number = 0
            self.last_bounding_box = BoundingBoxMeasures(token.bounding_box, self.last_measure_number,
                                                         self.last_measure_number)
            document.page_bounding_boxes[page_number] = self.last_bounding_box
        else:
            last_page_bb.bounding_box.extend(token.bounding_box)
            last_page_bb.to_measure = self.last_measure_number

    def import_file(self, file_path: Path) -> Document:
        """
        Import the content from the importer to the file.
        Args:
            file_path: The path to the file.

        Returns:
            Document - The document with the imported content.

        Examples:
            # Create the importer and read the file
            >>> importer = Importer()
            >>> importer.import_file('file.krn')
        """
        with open(file_path, 'r', newline='', encoding='utf-8', errors='ignore') as file:
            reader = csv.reader(file, delimiter='\t')
            return self.run(reader)

    def import_string(self, text: str) -> Document:
        """
        Import the content from the content of the score in string format.

        Args:
            text: The content of the score in string format.

        Returns:
            Document - The document with the imported content.

        Examples:
            # Create the importer and read the file
            >>> importer = Importer()
            >>> importer.import_string("**kern\n*clefF4\nc4\n4d\n4e\n4f\n*-")
            # Read the content from a file
            >>> with open('file.krn',  'r', newline='', encoding='utf-8', errors='ignore') as f: # We encourage you to use these open file options
            >>>     content = f.read()
            >>> importer.import_string(content)
            >>> document = importer.import_string(content)
        """
        lines = text.splitlines()
        reader = csv.reader(lines, delimiter='\t')
        return self.run(reader)

    def get_error_messages(self) -> str:
        """
        Get the error messages of the importer.

        Returns: str - The error messages split by a new line character.

        Examples:
            # Create the importer and read the file
            >>> importer = Importer()
            >>> importer.import_file(Path('file.krn'))
            >>> print(importer.get_error_messages())
            'Error: Invalid token in row 1'
        """
        result = ''
        for err in self.errors:
            result += str(err)
            result += '\n'
        return result

    def has_errors(self) -> bool:
        """
        Check if the importer has any errors.

        Returns: bool - True if the importer has errors, False otherwise.

        Examples:
            # Create the importer and read the file
            >>> importer = Importer()
            >>> importer.import_file(Path('file.krn'))    # file.krn has an error
            >>> print(importer.has_errors())
            True
            >>> importer.import_file(Path('file2.krn'))   # file2.krn has no errors
            >>> print(importer.has_errors())
            False
        """
        return len(self.errors) > 0

    def _compute_metacomment_token(self, raw_token: str):
        token = MetacommentToken(raw_token)
        if self._header_row_number is None:
            node = self._tree.add_node(self._tree_stage, self._last_node_previous_to_header, token, None, None, None)
            self._last_node_previous_to_header = node
        else:
            for parent in self._prev_stage_parents:
                node = self._tree.add_node(self._tree_stage, parent, token, self.get_last_spine_operator(parent), parent.last_signature_nodes, parent.header_node) # the same reference for all spines - TODO Recordar documentarlo
                self._next_stage_parents.append(node)

    def _compute_header_token(self, column_index: int, column_content: str):
        if self._header_row_number is not None and self._header_row_number != self._row_number:
            raise Exception(
                f"Several header rows not supported, there is a header row in #{self._header_row_number} and another in #{self._row_number} ")

            # it's a spine header
        self._document.header_stage = self._tree_stage
        importer = self._importers.get(column_content)
        if not importer:
            importer = createImporter(column_content)
            self._importers[column_content] = importer

        token = HeaderToken(column_content, spine_id=column_index)
        node = self._tree.add_node(self._tree_stage, self._last_node_previous_to_header, token, None, None)
        node.header_node = node # this value will be propagated
        self._next_stage_parents.append(node)

    def _compute_spine_operator_token(self, column_index: int, column_content: str, row: List[str]):
        token = SpineOperationToken(column_content)

        if column_index >= len(self._prev_stage_parents):
            raise Exception(f'Expected at least {column_index+1} parents in row {self._row_number}, but found {len(self._prev_stage_parents)}: {row}')

        parent = self._prev_stage_parents[column_index]
        node = self._tree.add_node(self._tree_stage, parent, token, self.get_last_spine_operator(parent), parent.last_signature_nodes, parent.header_node)

        if column_content == '*-':
            if node.last_spine_operator_node is not None:
                node.last_spine_operator_node.token.cancelled_at_stage = self._tree_stage
            pass # it's terminated, no continuation
        elif column_content == "*+" or column_content == "*^":
            self._next_stage_parents.append(node)
            self._next_stage_parents.append(node) # twice, the next stage two children will have this one as parent
        elif column_content == "*v":
            if node.last_spine_operator_node is not None:
                node.last_spine_operator_node.token.cancelled_at_stage = self._tree_stage

            if column_index == 0 or row[column_index-1] != '*v' or self._prev_stage_parents[column_index-1].header_node != self._prev_stage_parents[column_index].header_node: # don't collapse two different spines
                self._next_stage_parents.append(node) # just one spine each two
        else:
            raise Exception(f'Unknown spine operation in column #{column_content} and row #{self._row_number}')



"""
polish_scores

====

This module provides a way to download and process the Polish Scores dataset.
"""

from .download_polish_dataset import (
    convert_and_download_file,
    main as download_polish_scores,
)




"""
# kernpy

=====


Python Humdrum **kern and **mens utilities package.



Execute the following command to run **kernpy** as a module:
```shell
python -m kernpy --help
python -m kernpy <command> <options>
```
"""


from .core import *

from .io import *

from .util import *

from .polish_scores import *








from typing import Optional

from antlr4.error.ErrorListener import ConsoleErrorListener


class ParseError:
    def __init__(self, offendingSymbol, charPositionInLine, msg, exception):
        self.offendingSymbol = offendingSymbol
        self.charPositionInLine = charPositionInLine
        self.msg = msg
        self.exception = exception

    def __str__(self):
        return f"({self.charPositionInLine}): {self.msg}"

    def getOffendingSymbol(self):
        return self.offendingSymbol

    def getCharPositionInLine(self):
        return self.charPositionInLine

    def getMsg(self):
        return self.msg


class ErrorListener(ConsoleErrorListener):
    def __init__(self, *, verbose: Optional[bool] = False):
        """
        ErrorListener constructor.
        Args:
            verbose (bool): If True, the error messages will be printed to the console using \
            the `ConsoleErrorListener` interface.
        """
        super().__init__()
        self.errors = []
        self.verbose = verbose

    def syntaxError(self, recognizer, offendingSymbol, line, charPositionInLine, msg, e):
        if self.verbose:
            self.syntaxError(recognizer, offendingSymbol, line, charPositionInLine, msg, e)

        self.errors.append(ParseError(offendingSymbol, charPositionInLine, msg, e))

    def getNumberErrorsFound(self):
        return len(self.errors)

    def __str__(self):
        sb = ""
        for error in self.errors:
            sb += str(error) + "\n"
        return sb


import unittest
import os
from pathlib import Path

import kernpy as kp


class GkernTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.LINE_CHARACTER = 'L'
        cls.SPACE_CHARACTER = 'S'

    def test_line_character(self):
        self.assertEqual(kp.PositionInStaff.LINE_CHARACTER, self.LINE_CHARACTER)

    def test_space_character(self):
        self.assertEqual(kp.PositionInStaff.SPACE_CHARACTER, self.SPACE_CHARACTER)

    def test_from_line_bottom(self):
        p = kp.PositionInStaff.from_line(1)
        self.assertEqual(p, kp.PositionInStaff(0))

    def test_from_line_third(self):
        p = kp.PositionInStaff.from_line(3)
        self.assertEqual(p, kp.PositionInStaff(4))

    def test_from_space_bottom(self):
        p = kp.PositionInStaff.from_space(1)
        self.assertEqual(p, kp.PositionInStaff(1))

    def test_from_space_second(self):
        p = kp.PositionInStaff.from_space(2)
        self.assertEqual(p, kp.PositionInStaff(3))

    def test_from_encoded_line(self):
        p = kp.PositionInStaff.from_encoded(self.LINE_CHARACTER + '2')
        self.assertEqual(p, kp.PositionInStaff.from_line(2))

    def test_from_encoded_space(self):
        p = kp.PositionInStaff.from_encoded(self.SPACE_CHARACTER + '3')
        self.assertEqual(p, kp.PositionInStaff.from_space(3))

    def test_from_encoded_invalid(self):
        with self.assertRaises(ValueError):
            kp.PositionInStaff.from_encoded('X1')

    def test_line_number_1(self):
        p =kp. PositionInStaff(4)
        self.assertEqual(p.line(), 3.0)

    def test_line_number_2(self):
        p = kp.PositionInStaff(3)
        self.assertEqual(p.line(), 2)

    def test_line_number_3(self):
        p = kp.PositionInStaff(2)
        self.assertEqual(p.line(), 2.0)

    def test_line_number_4(self):
        p = kp.PositionInStaff(-5)
        self.assertEqual(p.line(), -2)

    def test_space_number_1(self):
        p = kp.PositionInStaff(3)
        self.assertEqual(p.space(), 2)

    def test_space_number_2(self):
        p = kp.PositionInStaff(1)
        self.assertEqual(p.space(), 1)

    def test_space_number_3(self):
        p = kp.PositionInStaff(2)
        self.assertEqual(p.space(), 1)

    def test_space_number_4(self):
        p = kp.PositionInStaff(5)
        self.assertEqual(p.space(), 3)

    def test_is_line_true(self):
        self.assertTrue(kp.PositionInStaff(0).is_line())
        self.assertTrue(kp.PositionInStaff(2).is_line())
        self.assertTrue(kp.PositionInStaff(4).is_line())
        self.assertTrue(kp.PositionInStaff(6).is_line())
        self.assertTrue(kp.PositionInStaff(8).is_line())
        self.assertTrue(kp.PositionInStaff(10).is_line())
        self.assertTrue(kp.PositionInStaff(-2).is_line())
        self.assertTrue(kp.PositionInStaff(-4).is_line())
        self.assertTrue(kp.PositionInStaff(-6).is_line())

    def test_is_line_false(self):
        self.assertFalse(kp.PositionInStaff(1).is_line())
        self.assertFalse(kp.PositionInStaff(3).is_line())
        self.assertFalse(kp.PositionInStaff(5).is_line())
        self.assertFalse(kp.PositionInStaff(7).is_line())
        self.assertFalse(kp.PositionInStaff(9).is_line())
        self.assertFalse(kp.PositionInStaff(11).is_line())
        self.assertFalse(kp.PositionInStaff(-1).is_line())
        self.assertFalse(kp.PositionInStaff(-3).is_line())
        self.assertFalse(kp.PositionInStaff(-5).is_line())


    def test_move_positive(self):
        p = kp.PositionInStaff(2).move(3)
        self.assertEqual(p, kp.PositionInStaff(5))

    def test_move_negative(self):
        p = kp.PositionInStaff(2).move(-2)
        self.assertEqual(p, kp.PositionInStaff(0))

    def test_position_above(self):
        p = kp.PositionInStaff(0).position_above()
        self.assertEqual(p, kp.PositionInStaff(2))

    def test_position_below(self):
        p = kp.PositionInStaff(2).position_below()
        self.assertEqual(p, kp.PositionInStaff(0))

    def test_str_from_line(self):
        p = kp.PositionInStaff.from_line(1)
        self.assertEqual(str(p), self.LINE_CHARACTER + '1')
        p = kp.PositionInStaff.from_line(2)
        self.assertEqual(str(p), self.LINE_CHARACTER + '2')
        p = kp.PositionInStaff.from_line(3)
        self.assertEqual(str(p), self.LINE_CHARACTER + '3')
        p = kp.PositionInStaff.from_line(4)
        self.assertEqual(str(p), self.LINE_CHARACTER + '4')
        p = kp.PositionInStaff.from_line(5)
        self.assertEqual(str(p), self.LINE_CHARACTER + '5')
        p = kp.PositionInStaff.from_line(-1)
        self.assertEqual(str(p), self.LINE_CHARACTER + '-1')
        p = kp.PositionInStaff.from_line(-2)
        self.assertEqual(str(p), self.LINE_CHARACTER + '-2')
        p = kp.PositionInStaff.from_line(-3)
        self.assertEqual(str(p), self.LINE_CHARACTER + '-3')

    def test_str_from_space(self):
        p = kp.PositionInStaff.from_space(1)
        self.assertEqual(str(p), self.SPACE_CHARACTER + '1')
        p = kp.PositionInStaff.from_space(2)
        self.assertEqual(str(p), self.SPACE_CHARACTER + '2')
        p = kp.PositionInStaff.from_space(3)
        self.assertEqual(str(p), self.SPACE_CHARACTER + '3')
        p = kp.PositionInStaff.from_space(4)
        self.assertEqual(str(p), self.SPACE_CHARACTER + '4')
        p = kp.PositionInStaff.from_space(5)
        self.assertEqual(str(p), self.SPACE_CHARACTER + '5')
        p = kp.PositionInStaff.from_space(-1)
        self.assertEqual(str(p), self.SPACE_CHARACTER + '-1')
        p = kp.PositionInStaff.from_space(-2)
        self.assertEqual(str(p), self.SPACE_CHARACTER + '-2')
        p = kp.PositionInStaff.from_space(-3)
        self.assertEqual(str(p), self.SPACE_CHARACTER + '-3')

    def test_equality(self):
        self.assertEqual(kp.PositionInStaff(2), kp.PositionInStaff(2))

    def test_inequality(self):
        self.assertNotEqual(kp.PositionInStaff(2), kp.PositionInStaff(3))

    def test_eq_different_type(self):
        self.assertFalse(kp.PositionInStaff(2) == 2)

    def test_lt(self):
        self.assertTrue(kp.PositionInStaff(1) < kp.PositionInStaff(2))

    def test_hash(self):
        s = {kp.PositionInStaff(2), kp.PositionInStaff(2)}
        self.assertEqual(len(s), 1)

    def test_from_line_ledger_line(self):
        p = kp.PositionInStaff.from_line(0)
        self.assertEqual(p, kp.PositionInStaff(-2))
        self.assertEqual(p.line(), 0.0)
        self.assertTrue(p.is_line())

    def test_from_space_ledger_space(self):
        p = kp.PositionInStaff.from_space(0)
        self.assertEqual(p, kp.PositionInStaff(-1))
        self.assertFalse(p.is_line())
        self.assertEqual(p.space(), 0)

    def test_str_ledger_line(self):
        p = kp.PositionInStaff.from_line(0)
        self.assertEqual(str(p), 'L0')

    def test_str_ledger_space(self):
        p = kp.PositionInStaff.from_space(0)
        self.assertEqual(str(p), 'S0')

    def test_is_line_negative_even(self):
        p = kp.PositionInStaff(-4)
        self.assertTrue(p.is_line())

    def test_is_line_negative_odd(self):
        p = kp.PositionInStaff(-3)
        self.assertFalse(p.is_line())

    def test_comparison_negative(self):
        low = kp.PositionInStaff(-3)
        high = kp.PositionInStaff(-1)
        self.assertTrue(low < high)


    def do_test_pitches_to_positions(self, reference: kp.PitchPositionReferenceSystem, pitch: kp.AgnosticPitch, expected_position: kp.PositionInStaff):
        computed_position = reference.compute_position(pitch)
        self.assertEqual(expected_position, computed_position)

    def test_compute_position(self):
        g_clef_reference = kp.PitchPositionReferenceSystem(kp.AgnosticPitch('E', 4))

        expected_position_in_staff = kp.PositionInStaff.from_line(1)
        computed_position_in_staff = g_clef_reference.compute_position(kp.AgnosticPitch('E', 4))

        self.assertEqual(expected_position_in_staff, computed_position_in_staff)

    def test_compute_position_bunch_of_samples_positives(self):
        g_clef_reference = kp.PitchPositionReferenceSystem(kp.AgnosticPitch('E', 4))

        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('E', 4), kp.PositionInStaff.from_line(1))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('F', 4), kp.PositionInStaff.from_space(1))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('G', 4), kp.PositionInStaff.from_line(2))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('A', 4), kp.PositionInStaff.from_space(2))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('B', 4), kp.PositionInStaff.from_line(3))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('C', 5), kp.PositionInStaff.from_space(3))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('D', 5), kp.PositionInStaff.from_line(4))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('E', 5), kp.PositionInStaff.from_space(4))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('F', 5), kp.PositionInStaff.from_line(5))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('G', 5), kp.PositionInStaff.from_space(5))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('A', 5), kp.PositionInStaff.from_line(6))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('B', 5), kp.PositionInStaff.from_space(6))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('C', 6), kp.PositionInStaff.from_line(7))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('D', 6), kp.PositionInStaff.from_space(7))

    def test_compute_position_bunch_of_samples_negatives(self):
        g_clef_reference = kp.PitchPositionReferenceSystem(kp.AgnosticPitch('E', 4))

        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('E', 4), kp.PositionInStaff.from_line(1))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('D', 4), kp.PositionInStaff.from_space(0))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('C', 4), kp.PositionInStaff.from_line(0))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('B', 3), kp.PositionInStaff.from_space(-1))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('A', 3), kp.PositionInStaff.from_line(-1))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('G', 3), kp.PositionInStaff.from_space(-2))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('F', 3), kp.PositionInStaff.from_line(-2))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('E', 3), kp.PositionInStaff.from_space(-3))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('D', 3), kp.PositionInStaff.from_line(-3))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('C', 3), kp.PositionInStaff.from_space(-4))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('B', 2), kp.PositionInStaff.from_line(-4))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('A', 2), kp.PositionInStaff.from_space(-5))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('G', 2), kp.PositionInStaff.from_line(-5))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('F', 2), kp.PositionInStaff.from_space(-6))

    def test_compute_position_bunch_of_samples_with_alterations(self):
        g_clef_reference = kp.PitchPositionReferenceSystem(kp.AgnosticPitch('E', 4))

        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('E', 4), kp.PositionInStaff.from_line(1))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('E+', 4), kp.PositionInStaff.from_line(1))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('E++', 4), kp.PositionInStaff.from_line(1))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('E-', 4), kp.PositionInStaff.from_line(1))
        self.do_test_pitches_to_positions(g_clef_reference, kp.AgnosticPitch('E--', 4), kp.PositionInStaff.from_line(1))

    def test_clef_ordinal_suffix(self):
        self.assertEqual(kp.Clef._ordinal_suffix(1), 'st')
        self.assertEqual(kp.Clef._ordinal_suffix(2), 'nd')
        self.assertEqual(kp.Clef._ordinal_suffix(3), 'rd')
        self.assertEqual(kp.Clef._ordinal_suffix(4), 'th')
        self.assertEqual(kp.Clef._ordinal_suffix(11), 'th')
        self.assertEqual(kp.Clef._ordinal_suffix(12), 'th')
        self.assertEqual(kp.Clef._ordinal_suffix(13), 'th')
        self.assertEqual(kp.Clef._ordinal_suffix(21), 'st')
        self.assertEqual(kp.Clef._ordinal_suffix(1_000_000_000_000_000_000_000), 'th')

    def test_g_clef_scale(self):
        gclef = kp.GClef()
        self.assertEqual(kp.AgnosticPitch('E', 4), gclef.bottom_line())
        self.do_test_pitches_to_positions(gclef.reference_point(), kp.AgnosticPitch('E', 4), kp.PositionInStaff.from_line(1))
        self.do_test_pitches_to_positions(gclef.reference_point(), kp.AgnosticPitch('F', 4), kp.PositionInStaff.from_space(1))
        self.do_test_pitches_to_positions(gclef.reference_point(), kp.AgnosticPitch('G', 4), kp.PositionInStaff.from_line(2))
        self.do_test_pitches_to_positions(gclef.reference_point(), kp.AgnosticPitch('A', 4), kp.PositionInStaff.from_space(2))
        self.do_test_pitches_to_positions(gclef.reference_point(), kp.AgnosticPitch('B', 4), kp.PositionInStaff.from_line(3))
        self.do_test_pitches_to_positions(gclef.reference_point(), kp.AgnosticPitch('C', 5), kp.PositionInStaff.from_space(3))
        self.do_test_pitches_to_positions(gclef.reference_point(), kp.AgnosticPitch('D', 5), kp.PositionInStaff.from_line(4))
        self.do_test_pitches_to_positions(gclef.reference_point(), kp.AgnosticPitch('E', 5), kp.PositionInStaff.from_space(4))

    def test_f3_clef_scale(self):
        f3clef = kp.F3Clef()
        self.assertEqual(kp.AgnosticPitch('B', 3), f3clef.bottom_line())
        self.do_test_pitches_to_positions(f3clef.reference_point(), kp.AgnosticPitch('B', 3), kp.PositionInStaff.from_line(1))
        self.do_test_pitches_to_positions(f3clef.reference_point(), kp.AgnosticPitch('C', 4), kp.PositionInStaff.from_space(1))
        self.do_test_pitches_to_positions(f3clef.reference_point(), kp.AgnosticPitch('D', 4), kp.PositionInStaff.from_line(2))
        self.do_test_pitches_to_positions(f3clef.reference_point(), kp.AgnosticPitch('E', 4), kp.PositionInStaff.from_space(2))
        self.do_test_pitches_to_positions(f3clef.reference_point(), kp.AgnosticPitch('F', 4), kp.PositionInStaff.from_line(3))
        self.do_test_pitches_to_positions(f3clef.reference_point(), kp.AgnosticPitch('G', 4), kp.PositionInStaff.from_space(3))
        self.do_test_pitches_to_positions(f3clef.reference_point(), kp.AgnosticPitch('A', 4), kp.PositionInStaff.from_line(4))
        self.do_test_pitches_to_positions(f3clef.reference_point(), kp.AgnosticPitch('B', 4), kp.PositionInStaff.from_space(4))

    def test_f4_clef_scale(self):
        f4clef = kp.F4Clef()
        self.assertEqual(kp.AgnosticPitch('G', 2), f4clef.bottom_line())
        self.do_test_pitches_to_positions(f4clef.reference_point(), kp.AgnosticPitch('G', 2), kp.PositionInStaff.from_line(1))
        self.do_test_pitches_to_positions(f4clef.reference_point(), kp.AgnosticPitch('A', 2), kp.PositionInStaff.from_space(1))
        self.do_test_pitches_to_positions(f4clef.reference_point(), kp.AgnosticPitch('B', 2), kp.PositionInStaff.from_line(2))
        self.do_test_pitches_to_positions(f4clef.reference_point(), kp.AgnosticPitch('C', 3), kp.PositionInStaff.from_space(2))
        self.do_test_pitches_to_positions(f4clef.reference_point(), kp.AgnosticPitch('D', 3), kp.PositionInStaff.from_line(3))
        self.do_test_pitches_to_positions(f4clef.reference_point(), kp.AgnosticPitch('E', 3), kp.PositionInStaff.from_space(3))
        self.do_test_pitches_to_positions(f4clef.reference_point(), kp.AgnosticPitch('F', 3), kp.PositionInStaff.from_line(4))
        self.do_test_pitches_to_positions(f4clef.reference_point(), kp.AgnosticPitch('G', 3), kp.PositionInStaff.from_space(4))

    def test_c1_clef_scale(self):
        c1clef = kp.C1Clef()
        self.assertEqual(kp.AgnosticPitch('C', 3), c1clef.bottom_line())
        self.do_test_pitches_to_positions(c1clef.reference_point(), kp.AgnosticPitch('C', 3), kp.PositionInStaff.from_line(1))
        self.do_test_pitches_to_positions(c1clef.reference_point(), kp.AgnosticPitch('D', 3), kp.PositionInStaff.from_space(1))
        self.do_test_pitches_to_positions(c1clef.reference_point(), kp.AgnosticPitch('E', 3), kp.PositionInStaff.from_line(2))
        self.do_test_pitches_to_positions(c1clef.reference_point(), kp.AgnosticPitch('F', 3), kp.PositionInStaff.from_space(2))
        self.do_test_pitches_to_positions(c1clef.reference_point(), kp.AgnosticPitch('G', 3), kp.PositionInStaff.from_line(3))
        self.do_test_pitches_to_positions(c1clef.reference_point(), kp.AgnosticPitch('A', 3), kp.PositionInStaff.from_space(3))
        self.do_test_pitches_to_positions(c1clef.reference_point(), kp.AgnosticPitch('B', 3), kp.PositionInStaff.from_line(4))
        self.do_test_pitches_to_positions(c1clef.reference_point(), kp.AgnosticPitch('C', 4), kp.PositionInStaff.from_space(4))

    def test_c2_clef_scale(self):
        c2clef = kp.C2Clef()
        self.assertEqual(kp.AgnosticPitch('A', 2), c2clef.bottom_line())
        self.do_test_pitches_to_positions(c2clef.reference_point(), kp.AgnosticPitch('A', 2), kp.PositionInStaff.from_line(1))
        self.do_test_pitches_to_positions(c2clef.reference_point(), kp.AgnosticPitch('B', 2), kp.PositionInStaff.from_space(1))
        self.do_test_pitches_to_positions(c2clef.reference_point(), kp.AgnosticPitch('C', 2), kp.PositionInStaff.from_line(2))
        self.do_test_pitches_to_positions(c2clef.reference_point(), kp.AgnosticPitch('D', 2), kp.PositionInStaff.from_space(2))
        self.do_test_pitches_to_positions(c2clef.reference_point(), kp.AgnosticPitch('E', 2), kp.PositionInStaff.from_line(3))
        self.do_test_pitches_to_positions(c2clef.reference_point(), kp.AgnosticPitch('F', 2), kp.PositionInStaff.from_space(3))
        self.do_test_pitches_to_positions(c2clef.reference_point(), kp.AgnosticPitch('G', 2), kp.PositionInStaff.from_line(4))
        self.do_test_pitches_to_positions(c2clef.reference_point(), kp.AgnosticPitch('A', 3), kp.PositionInStaff.from_space(4))

    def test_c3_clef_scale(self):
        c3clef = kp.C3Clef()
        self.assertEqual(kp.AgnosticPitch('B', 2), c3clef.bottom_line())
        self.do_test_pitches_to_positions(c3clef.reference_point(), kp.AgnosticPitch('B', 2), kp.PositionInStaff.from_line(1))
        self.do_test_pitches_to_positions(c3clef.reference_point(), kp.AgnosticPitch('C', 2), kp.PositionInStaff.from_space(1))
        self.do_test_pitches_to_positions(c3clef.reference_point(), kp.AgnosticPitch('D', 2), kp.PositionInStaff.from_line(2))
        self.do_test_pitches_to_positions(c3clef.reference_point(), kp.AgnosticPitch('E', 2), kp.PositionInStaff.from_space(2))
        self.do_test_pitches_to_positions(c3clef.reference_point(), kp.AgnosticPitch('F', 2), kp.PositionInStaff.from_line(3))
        self.do_test_pitches_to_positions(c3clef.reference_point(), kp.AgnosticPitch('G', 2), kp.PositionInStaff.from_space(3))
        self.do_test_pitches_to_positions(c3clef.reference_point(), kp.AgnosticPitch('A', 2), kp.PositionInStaff.from_line(4))
        self.do_test_pitches_to_positions(c3clef.reference_point(), kp.AgnosticPitch('B', 3), kp.PositionInStaff.from_space(4))

    def test_c4_clef_scale(self):
        c4clef = kp.C4Clef()
        self.assertEqual(kp.AgnosticPitch('D', 2), c4clef.bottom_line())
        self.do_test_pitches_to_positions(c4clef.reference_point(), kp.AgnosticPitch('D', 2), kp.PositionInStaff.from_line(1))
        self.do_test_pitches_to_positions(c4clef.reference_point(), kp.AgnosticPitch('E', 2), kp.PositionInStaff.from_space(1))
        self.do_test_pitches_to_positions(c4clef.reference_point(), kp.AgnosticPitch('F', 2), kp.PositionInStaff.from_line(2))
        self.do_test_pitches_to_positions(c4clef.reference_point(), kp.AgnosticPitch('G', 2), kp.PositionInStaff.from_space(2))
        self.do_test_pitches_to_positions(c4clef.reference_point(), kp.AgnosticPitch('A', 2), kp.PositionInStaff.from_line(3))
        self.do_test_pitches_to_positions(c4clef.reference_point(), kp.AgnosticPitch('B', 2), kp.PositionInStaff.from_space(3))
        self.do_test_pitches_to_positions(c4clef.reference_point(), kp.AgnosticPitch('C', 3), kp.PositionInStaff.from_line(4))
        self.do_test_pitches_to_positions(c4clef.reference_point(), kp.AgnosticPitch('D', 3), kp.PositionInStaff.from_space(4))


class TestClefFactory(unittest.TestCase):
    def test_G_clef_without_decorators(self):
        clef = kp.ClefFactory.create_clef('*clefG2')
        self.assertIsInstance(clef, kp.GClef)

    def test_G_clef_single_v(self):
        clef = kp.ClefFactory.create_clef('*clefGv2')
        self.assertIsInstance(clef, kp.GClef)

    def test_G_clef_double_vv(self):
        clef = kp.ClefFactory.create_clef('*clefGvv2')
        self.assertIsInstance(clef, kp.GClef)

    def test_G_clef_single_caret(self):
        clef = kp.ClefFactory.create_clef('*clefG^2')
        self.assertIsInstance(clef, kp.GClef)

    def test_G_clef_double_caret(self):
        clef = kp.ClefFactory.create_clef('*clefG^^^^2')
        self.assertIsInstance(clef, kp.GClef)

    # F‑clef (bass), line 3
    def test_F3_clef_without_decorators(self):
        clef = kp.ClefFactory.create_clef('*clefF3')
        self.assertIsInstance(clef, kp.F3Clef)

    def test_F3_clef_with_v(self):
        clef = kp.ClefFactory.create_clef('*clefFv3')
        self.assertIsInstance(clef, kp.F3Clef)

    def test_F4_clef_without_decorators(self):
        clef = kp.ClefFactory.create_clef('*clefF4')
        self.assertIsInstance(clef, kp.F4Clef)

    def test_F4_clef_with_caret(self):
        clef = kp.ClefFactory.create_clef('*clefF^4')
        self.assertIsInstance(clef, kp.F4Clef)

    # C‑clefs, lines 1–4
    def test_C1_clef_basic(self):
        clef = kp.ClefFactory.create_clef('*clefC1')
        self.assertIsInstance(clef, kp.C1Clef)

    def test_C2_clef_with_vv(self):
        clef = kp.ClefFactory.create_clef('*clefCvv2')
        self.assertIsInstance(clef, kp.C2Clef)

    def test_C3_clef_with_caret(self):
        clef = kp.ClefFactory.create_clef('*clefC^3')
        self.assertIsInstance(clef, kp.C3Clef)

    def test_C4_clef_with_double_caret(self):
        clef = kp.ClefFactory.create_clef('*clefC^^4')
        self.assertIsInstance(clef, kp.C4Clef)

    # invalid name
    def test_invalid_clef_name(self):
        with self.assertRaises(ValueError):
            kp.ClefFactory.create_clef('*clefH2')

    # invalid F‑clef line
    def test_invalid_F_line(self):
        with self.assertRaises(ValueError):
            kp.ClefFactory.create_clef('*clefF2')

    # invalid C‑clef line
    def test_invalid_C_line(self):
        with self.assertRaises(ValueError):
            kp.ClefFactory.create_clef('*clefC5')

    # missing leading *
    def test_missing_prefix(self):
        with self.assertRaises(ValueError):
            kp.ClefFactory.create_clef('clefG2')
        with self.assertRaises(ValueError):
            kp.ClefFactory.create_clef('random')






import unittest
from unittest.mock import Mock, MagicMock

import kernpy as kp

class TestStoreCache(unittest.TestCase):
    def test_request_calculates_value(self):
        """Test that the callback is called on the first request."""
        callback = Mock(return_value=10)
        cache = kp.util.store_cache.StoreCache()

        # First request: callback should be called.
        result = cache.request(callback, 5)
        self.assertEqual(result, 10)
        callback.assert_called_once_with(5)

    def test_request_uses_cache(self):
        """Test that subsequent requests use the cached value and do not call the callback again."""
        callback = Mock(return_value=10)
        cache = kp.util.store_cache.StoreCache()

        # First call to store the value in cache.
        first_result = cache.request(callback, 5)
        self.assertEqual(first_result, 10)
        callback.assert_called_once_with(5)

        # Reset the mock call history.
        callback.reset_mock()

        # Second call should retrieve value from cache.
        second_result = cache.request(callback, 5)
        self.assertEqual(second_result, 10)
        callback.assert_not_called()  # The callback should not have been called again.

    def test_hierarchy_nodes(self):
        expected_nodes = {kp.TokenCategory.PITCH, kp.TokenCategory.ALTERATION, kp.TokenCategory.DECORATION}

        # 1. Assert that the direct call returns the expected nodes.
        nodes_result = kp.TokenCategoryHierarchyMapper.nodes(kp.TokenCategory.NOTE)
        self.assertEqual(nodes_result, expected_nodes)

        # 2. Create an instance of the cache and request the nodes value via the cache.
        cache = kp.util.store_cache.StoreCache()
        response1 = cache.request(kp.TokenCategoryHierarchyMapper.nodes, kp.TokenCategory.NOTE)
        self.assertEqual(response1, expected_nodes)

        # 3. Request again to confirm that the cached value is used (the result remains the same).
        response2 = cache.request(kp.TokenCategoryHierarchyMapper.nodes, kp.TokenCategory.NOTE)
        self.assertEqual(response2, expected_nodes)

        # 4. Assert that the callback was only invoked once,
        # you can wrap the nodes method with a MagicMock. For example:
        original_nodes = kp.TokenCategoryHierarchyMapper.nodes
        kp.TokenCategoryHierarchyMapper.nodes = MagicMock(wraps=original_nodes)

        # Call through the cache
        cache = kp.util.store_cache.StoreCache()
        cache.request(kp.TokenCategoryHierarchyMapper.nodes, kp.TokenCategory.NOTE)
        cache.request(kp.TokenCategoryHierarchyMapper.nodes, kp.TokenCategory.NOTE)

        # Assert that the original method was only called once.
        kp.TokenCategoryHierarchyMapper.nodes.assert_called_once_with(kp.TokenCategory.NOTE)

        # Restore the original nodes method.
        kp.TokenCategoryHierarchyMapper.nodes = original_nodes


if __name__ == '__main__':
    unittest.main()



import unittest

import kernpy as kp


class TestExportOptions(unittest.TestCase):
    def test_default_values(self):
        expected_options = kp.ExportOptions()
        expected_options.spine_types = kp.core.HEADERS
        expected_options.token_categories = [c for c in kp.TokenCategory]
        expected_options.from_measure = None
        expected_options.to_measure = None
        expected_options.kern_type = kp.Encoding.normalizedKern
        expected_options.instruments = None
        expected_options.show_measure_numbers = False
        expected_options.spine_ids = None

        real_options = kp.ExportOptions()

        self.assertEqual(expected_options, real_options)

    def test__eq__(self):
        a = kp.ExportOptions()
        b = kp.ExportOptions()
        self.assertEqual(a, b)

    def test__ne__(self):
        a = kp.ExportOptions()
        b = kp.ExportOptions(instruments=['piano'])
        self.assertNotEqual(a, b)





import unittest
from unittest.mock import MagicMock, PropertyMock, patch

import kernpy as kp


class TestTokenizer(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.token_1 = MagicMock(spec=kp.Token)  # Mock Token class

        # Mock the encoding property
        type(cls.token_1).encoding = PropertyMock(return_value="2.bb-_L")

        # Mock the export function
        cls.token_1.export = MagicMock(return_value="2@.@bb@-·_·L")

        cls.default_categories = set([c for c in kp.TokenCategory])



    def test_ekern_tokenizer_1(self):
        tokenizer = kp.EkernTokenizer(token_categories=self.default_categories)
        token_str = tokenizer.tokenize(self.token_1)
        self.assertEqual('2@.@bb@-·_·L', token_str)
        self.token_1.export.assert_called()

    def test_kern_tokenizer_1(self):
        tokenizer = kp.KernTokenizer(token_categories=self.default_categories)
        token_str = tokenizer.tokenize(self.token_1)
        self.assertEqual('2.bb-_L', token_str)
        self.token_1.export.assert_called()

    def test_bkern_tokenizer_1(self):
        tokenizer = kp.BkernTokenizer(token_categories=self.default_categories)
        token_str = tokenizer.tokenize(self.token_1)
        self.assertEqual('2.bb-', token_str)
        self.token_1.export.assert_called()

    def test_bekern_tokenizer_1(self):
        tokenizer = kp.BekernTokenizer(token_categories=self.default_categories)
        token_str = tokenizer.tokenize(self.token_1)
        self.assertEqual('2@.@bb@-', token_str)
        self.token_1.export.assert_called()

    def test_tokenizer_factory_kern(self):
        tokenizer = kp.TokenizerFactory.create(kp.Encoding.normalizedKern.value, token_categories=self.default_categories)
        self.assertIsInstance(tokenizer, kp.KernTokenizer)

    def test_tokenizer_factory_ekern(self):
        tokenizer = kp.TokenizerFactory.create(kp.Encoding.eKern.value, token_categories=self.default_categories)
        self.assertIsInstance(tokenizer, kp.EkernTokenizer)

    def test_tokenizer_factory_bkern(self):
        tokenizer = kp.TokenizerFactory.create(kp.Encoding.bKern.value, token_categories=self.default_categories)
        self.assertIsInstance(tokenizer, kp.BkernTokenizer)


    def test_tokenizer_factory_basic_kern(self):
        tokenizer = kp.TokenizerFactory.create(kp.Encoding.basicKern.value, token_categories=self.default_categories)
        self.assertIsInstance(tokenizer, kp.BkernTokenizer)

    def test_tokenizer_factory_basic_extended_kern(self):
        tokenizer = kp.TokenizerFactory.create(kp.Encoding.basicExtendedKern.value, token_categories=self.default_categories)
        self.assertIsInstance(tokenizer, kp.BekernTokenizer)

    def test_tokenizer_factory_bekern(self):
        tokenizer = kp.TokenizerFactory.create(kp.Encoding.bEkern.value, token_categories=self.default_categories)
        self.assertIsInstance(tokenizer, kp.BekernTokenizer)

    def test_tokenizer_factory_agnostic_kern(self):
        tokenizer = kp.TokenizerFactory.create(kp.Encoding.agnosticKern.value, token_categories=self.default_categories)
        self.assertIsInstance(tokenizer, kp.AKernTokenizer)

    def test_tokenizer_factory_agnostic_extended_kern(self):
        tokenizer = kp.TokenizerFactory.create(kp.Encoding.agnosticExtendedKern.value, token_categories=self.default_categories)
        self.assertIsInstance(tokenizer, kp.AEKernTokenizer)

    def test_tokenizer_factory_raise_error_if_none(self):
        with self.assertRaises(ValueError):
            kp.TokenizerFactory.create(None, token_categories=self.default_categories)

    def test_tokenizer_factory_raise_error_if_invalid(self):
        with self.assertRaises(ValueError):
            kp.TokenizerFactory.create('invalid', token_categories=self.default_categories)



import csv
import string
import logging
from collections.abc import Iterable
from enum import Enum

from .importer_factory import createImporter
from .tokens import HeaderToken, SpineOperationToken, TokenCategory, BoundingBoxToken, KeySignatureToken, \
    TimeSignatureToken, MeterSymbolToken, ClefToken, BarToken, MetacommentToken, ErrorToken, FieldCommentToken, \
    BEKERN_CATEGORIES, TOKEN_SEPARATOR, DECORATION_SEPARATOR


class Encoding(Enum):  # TODO: Eventually, polymorphism will be used to export different types of kern files
    """
    Options for exporting a kern file.

    Example:
        # Create the importer
        >>> hi = HumdrumImporter()

        # Read the file
        >>> hi.import_file('file.krn')

        # Export the file
        >>> options = ExportOptions(spine_types=['**kern'], token_categories=BEKERN_CATEGORIES, kernType=Encoding.normalizedKern)
        >>> exported = hi.doExport(options)

    """
    unprocessed = 0
    eKern = 1
    normalizedKern = 2


class ExportOptions:
    def __init__(self, spine_types=None, token_categories=None, from_measure: int = None, to_measure: int = None, kern_type: Encoding = Encoding.normalizedKern, instruments=None):
        """
        Create a new ExportOptions object.

        Args:
            spine_types (Iterable): **kern, **mens, etc...
            token_categories (Iterable): TokenCategory
            from_measure (int): The measure to start exporting. When None, the exporter will start from the beginning of the file.
            to_measure (int): The measure to end exporting. When None, the exporter will end at the end of the file.
            kern_type (Encoding): The type of the kern file to export.
            instruments (Iterable): The instruments to export. When None, all the instruments will be exported.


        Example:
            >>> from kernpy import HumdrumImporter, ExportOptions

            Create the importer and read the file
            >>> hi = HumdrumImporter()
            >>> hi.import_file('file.krn')

            Export the file with the specified options
            >>> options = ExportOptions(spine_types=['**kern'], token_categories=BEKERN_CATEGORIES)
            >>> exported_data = hi.doExport(options)

            Export only the lyirics
            >>> options = ExportOptions(spine_types=['**kern'], token_categories=[TokenCategory.LYRICS])
            >>> exported_data = hi.doExport(options)

            Export the comments
            >>> options = ExportOptions(spine_types=['**kern'], token_categories=[TokenCategory.LINE_COMMENTS, TokenCategory.FIELD_COMMENTS])
            >>> exported_data = hi.doExport(options)

            Export using the eKern version
            >>> options = ExportOptions(spine_types=['**kern'], token_categories=BEKERN_CATEGORIES, kern_type=Encoding.eKern)
            >>> exported_data = hi.doExport(options)

        """
        self.spine_types = spine_types or []
        self.from_measure = from_measure
        self.to_measure = to_measure
        self.token_categories = token_categories or []
        self.kern_type = kern_type
        self.instruments = instruments or []


class BoundingBoxMeasures:
    def __init__(self, bounding_box, from_measure, to_measure):
        self.from_measure = from_measure
        self.to_measure = to_measure
        self.bounding_box = bounding_box


class Spine:
    def __init__(self, spine_type, importer):
        self.spine_type = spine_type  # **mens, **kern, etc...
        self.rows = []  # each row will contain just one item or an array of items of type Token
        self.importer = importer
        self.importing_subspines = 1  # 0 for terminated subspines - used just for importing
        self.next_row_subspine_variation = 0  # when a spine operation is added or removed, the subspines number must be modified for the next row

    def size(self):
        return len(self.rows)

    def __len__(self):
        return self.size()

    def isTerminated(self):
        return self.importing_subspines > 0

    def getNumSubspines(self, row_number):
        if row_number < 0:
            raise Exception(f'Negative row number {row_number}')
        if row_number >= len(self.rows):
            raise Exception(f'Row number {row_number} out of bounds {len(self.rows)}')

        return len(self.rows[row_number])

    def addRow(self):
        if self.importing_subspines != 0:  # if not terminated
            self.rows.append([])
            if self.next_row_subspine_variation > 0:
                new_subspines = self.importing_subspines + self.next_row_subspine_variation
            elif self.next_row_subspine_variation < 0:
                new_subspines = self.importing_subspines + (
                        self.next_row_subspine_variation + 1)  # e.g. *v *v *v for three spines lead to 1 spine
            else:
                new_subspines = self.importing_subspines
            logging.debug(f'Adding row to spine, previous subspines={self.importing_subspines}, new={new_subspines}')
            self.importing_subspines = new_subspines
            self.next_row_subspine_variation = 0

    def addToken(self, token):
        if not token:
            raise Exception('Trying to add a empty token')

        row = len(self.rows) - 1
        if not isinstance(token, MetacommentToken) and len(self.rows[row]) >= self.importing_subspines:
            raise Exception(
                f'There are already {len(self.rows[row])} subspines, and this spine should have at most {self.importing_subspines}')

        self.rows[row].append(token)

    def increaseSubspines(self):
        self.next_row_subspine_variation = self.next_row_subspine_variation + 1

    def decreaseSubspines(self):
        self.next_row_subspine_variation = self.next_row_subspine_variation - 1

    def terminate(self):
        self.importing_subspines = 0

    def isFullRow(self):
        if self.importing_subspines == 0:
            return True
        else:
            row = len(self.rows) - 1
            return len(self.rows[row]) >= self.importing_subspines

    def isContentOfType(self, row, clazz):
        self.checkRowIndex(row)
        for subspine in self.rows[row]:
            if isinstance(subspine, clazz):
                return True
        return False

    def checkRowIndex(self, row):
        if row < 0:
            raise Exception(f'Negative row {row}')
        if row >= len(self.rows):
            raise Exception(f'Row {row} out of bounds {len(self.rows)}')

    def getRowContent(self, row, kern_type: Encoding, token_categories: Iterable) -> string:
        self.checkRowIndex(row)

        result = ''
        for subspine in self.rows[row]:
            if subspine.category == TokenCategory.STRUCTURAL or subspine.category in token_categories:
                if len(result) > 0:
                    result += '\t'
                if kern_type == Encoding.unprocessed:
                    result += subspine.encoding
                elif kern_type in {Encoding.eKern, Encoding.normalizedKern}:
                    if subspine.hidden:
                        exp = '.'
                    else:
                        exp = subspine.export()
                        if kern_type == Encoding.normalizedKern:
                            exp = get_kern_from_ekern(exp)
                    if not exp:
                        raise Exception(f'Subspine {subspine.encoding} is exported as None')
                    result += exp
                else:
                    raise ValueError(f'Unknown kern type {kern_type}.\nView {help(Encoding)} ')

        return result


class Signatures:
    def __init__(self, header_row, clef_row, key_signature_row, time_signature_row, meter_symbol_row):
        self.last_header_row = header_row
        self.last_clef_row = clef_row
        self.last_key_signature_row = key_signature_row
        self.last_time_signature_row = time_signature_row
        self.last_meter_symbol_row = meter_symbol_row

    def clone(self):
        return Signatures(self.last_header_row, self.last_clef_row, self.last_key_signature_row,
                          self.last_time_signature_row, self.last_meter_symbol_row)


class HumdrumImporter:
    HEADERS = {"**mens", "**kern", "**text", "**harm", "**mxhm", "**root", "**dyn", "**dynam", "**fing"}
    SPINE_OPERATIONS = {"*-", "*+", "*^", "*v"}

    def __init__(self):
        self.spines = []
        self.current_spine_index = 0
        # self.page_start_rows = []
        self.measure_start_rows = []  # starting from 1. Rows after removing empty lines and line comments
        self.page_bounding_boxes = {}
        self.last_measure_number = None
        self.last_bounding_box = None
        self.errors = []

    def getMetacomments(self, KeyComment: str = None, clean: bool = True):  # each metacomment is contained in all spines as a reference to the same object
        """
        Get the metacomments of the file.

        Args:
            KeyComment: The key of the metacomment. (optional).\
                If not specified, all the metacomments will be returned.
                If specified, all the content of the metacomment with the specified key will be returned.
            clean: If True, the metacomments will be returned applying a .strip(). Only valid if KeyComment is not None.

        Returns:
            A list with the metacomments.\
                if KeyComment is not None, a list be returned anyway. \
                If there are no metacomments with the specified key, an empty list will be returned.

        Example:
            >>> from kernpy import HumdrumImporter
            >>> importer = HumdrumImporter()

            # Read the file
            >>> importer.import_file('file.krn')

            # Get all the metacomments
            >>> all_metacomments = importer.getMetacomments()
            # ... modify the metacomments using your own logic

            # Get the metacomments with the key: get the composer:
            >>> composer = importer.getMetacomments(KeyComment='!!!COM')

            # check if your kern file format is compatible with the expected format. If it is not, do not clen it:
            >>> raw_compose = importer.getMetacomments(KeyComment='!!!COM', clean=False)

        """
        result = []
        for token in self.spines[0].rows:
            if isinstance(token[0], MetacommentToken):
                if clean:
                    result.append(token[0].encoding.strip())
                else:
                    result.append(token[0].encoding)

        if KeyComment is not None:
            clean_rows = [row.replace('!!!', '').replace('!!', '') for row in result]
            filtered_rows = [row for row in clean_rows if row.startswith(KeyComment)]
            valid_rows = [row.replace(KeyComment, '').strip()[2:] for row in filtered_rows] if clean else filtered_rows
            return valid_rows

        return result

    def doImport(self, reader):
        importers = {}
        header_row_number = None
        row_number = 1
        pending_metacomments = []  # those appearing before the headers
        for row in reader:
            for spine in self.spines:
                self.current_spine_index = 0
                spine.addRow()
            if len(row) > 0:  # the last one
                if row[0].startswith("!!"):
                    mt = MetacommentToken(row[0])
                    if len(self.spines) == 0:
                        pending_metacomments.append(mt)
                    else:
                        for spine in self.spines:
                            spine.addToken(mt)  # the same reference for all spines
                else:
                    is_barline = False
                    for column in row:
                        if column in self.HEADERS:
                            if header_row_number is not None and header_row_number != row_number:
                                raise Exception(
                                    f"Several header rows not supported, there is a header row in #{header_row_number} and another in #{row_number} ")

                            header_row_number = row_number
                            importer = importers.get(column)
                            if not importer:
                                importer = createImporter(column)
                                importers[column] = importer
                            spine = Spine(column, importer) # TODO: Add instrument
                            for pending_metacomment in pending_metacomments:
                                spine.addRow()
                                spine.addToken(pending_metacomment)  # same reference for all spines

                            token = HeaderToken(column)
                            spine.addRow()
                            spine.addToken(token)
                            self.spines.append(spine)
                        else:
                            try:
                                current_spine = self.getNextSpine()
                                logging.debug(
                                    f'Row #{row_number}, current spine #{self.current_spine_index} of size {current_spine.importing_subspines}, and importer {current_spine.importer}')
                            except Exception as e:
                                raise Exception(
                                    f'Cannot get next spine at row {row_number}: {e} while reading row {row} ')

                            if column in self.SPINE_OPERATIONS:
                                current_spine.addToken(SpineOperationToken(column))

                                if column == '*-':
                                    current_spine.terminate()
                                elif column == "*+" or column == "*^":
                                    current_spine.increaseSubspines()
                                elif column == "*v":
                                    current_spine.decreaseSubspines()
                            else:
                                if column.startswith("!"):
                                    token = FieldCommentToken(column)
                                else:
                                    try:
                                        token = current_spine.importer.run(column)
                                    except Exception as error:
                                        token = ErrorToken(column, row_number, error)
                                        self.errors.append(token)
                                if not token:
                                    raise Exception(
                                        f'No token generated for input {column} in row number #{row_number} using importer {current_spine.importer}')
                                current_spine.addToken(token)
                                if token.category == TokenCategory.BARLINES or token.category == TokenCategory.CORE and len(
                                        self.measure_start_rows) == 0:
                                    is_barline = True
                                elif isinstance(token, BoundingBoxToken):
                                    self.handleBoundingBox(token)

                    if is_barline:
                        self.measure_start_rows.append(row_number)
                        self.last_measure_number = len(self.measure_start_rows)
                        if self.last_bounding_box:
                            self.last_bounding_box.to_measure = self.last_measure_number
                row_number = row_number + 1

    def doImportFile(self, file_path: string):
        """
        Import the content from the importer to the file.
        Args:
            file_path: The path to the file.

        Returns:
            None

        Example:
            # Create the importer and read the file
            >>> hi = HumdrumImporter()
            >>> hi.import_file('file.krn')
        """
        with open(file_path, 'r', newline='', encoding='utf-8', errors='ignore') as file:
            reader = csv.reader(file, delimiter='\t')
            self.doImport(reader)

    def doImportString(self, text: string):
        lines = text.splitlines()
        reader = csv.reader(lines)
        self.doImport(reader)

    def getSpine(self, index: int) -> Spine:
        if index < 0:
            raise Exception(f'Negative index {index}')
        elif index >= len(self.spines):
            raise Exception(f'Index {index} out of bounds for an array of {len(self.spines)} spines')
        return self.spines[index]

    def getNextSpine(self):
        spine = self.getSpine(self.current_spine_index)
        while spine.isFullRow() and self.current_spine_index < (len(self.spines) - 1):
            self.current_spine_index = self.current_spine_index + 1
            spine = self.getSpine(self.current_spine_index)

        if self.current_spine_index == len(self.spines):
            raise Exception('All spines are full, the spine divisions may be wrong')

        return spine

    def doExportNormalizedKern(self, options: ExportOptions) -> string:
        options.kern_type = Encoding.normalizedKern
        return self.doExport(options)

    def doExportEKern(self, options: ExportOptions) -> string:
        options.kern_type = Encoding.eKern
        return self.doExport(options)

    def doExportUnprocessed(self, options: ExportOptions) -> string:
        options.kern_type = Encoding.unprocessed
        return self.doExport(options)

    def handleBoundingBox(self, token: BoundingBoxToken):
        page_number = token.page_number
        last_page_bb = self.page_bounding_boxes.get(page_number)
        if last_page_bb is None:
            # print(f'Adding {page_number}')
            if self.last_measure_number is None:
                self.last_measure_number = 0
            self.last_bounding_box = BoundingBoxMeasures(token.bounding_box, self.last_measure_number,
                                                         self.last_measure_number)
            self.page_bounding_boxes[page_number] = self.last_bounding_box
        else:
            # print(f'Extending page {page_number}')
            last_page_bb.bounding_box.extend(token.bounding_box)
            last_page_bb.to_measure = self.last_measure_number

    def getMaxRows(self):
        return max(spine.size() for spine in self.spines)

    def checkMeasure(self, measure_number):
        if measure_number < 0:
            raise Exception(f'The measure number must be >=1, and it is {measure_number}')

        max_measures = len(self.measure_start_rows)
        if measure_number > max_measures:
            raise Exception(f'The measure number must be <= {max_measures}, and it is {measure_number}')

    def doExport(self, options: ExportOptions) -> string:
        max_rows = self.getMaxRows()
        signatures_at_each_row = []
        row_contents = []

        if options.from_measure is not None and options.from_measure < 0:
            raise ValueError(f'option from_measure must be >=0 but {options.from_measure} was found. ')
        if options.to_measure is not None and options.to_measure > len(self.measure_start_rows):
            #"TODO: DAVID, check options.to_measure bounds. len(self.measure_start_rows) or len(self.measure_start_rows) - 1"
            raise ValueError(f'option to_measure must be <= {len(self.measure_start_rows)} but {options.to_measure} was found. ')
        if options.to_measure is not None and options.from_measure is not None and options.to_measure < options.from_measure:
            raise ValueError(f'option to_measure must be >= from_measure but {options.to_measure} < {options.from_measure} was found. ')

        last_signature = None
        for i in range(max_rows):
            row_result = ''
            if last_signature:
                current_signature = last_signature.clone()
            else:
                current_signature = Signatures(None, None, None, None, None)
            last_signature = current_signature
            empty = True
            for spine in self.spines:
                if spine.spine_type in options.spine_types:
                    if i < spine.size():  # required because the spine may be terminated
                        if len(row_result) > 0:
                            row_result += '\t'

                        content = spine.getRowContent(i, options.kern_type, options.token_categories)

                        if content and content != '.' and content != '*':
                            empty = False
                            if options.from_measure:  # if not, we don't need to compute this value
                                if spine.isContentOfType(i, HeaderToken):
                                    current_signature.last_header_row = i
                                elif spine.isContentOfType(i, ClefToken):
                                    current_signature.last_clef_row = i
                                elif spine.isContentOfType(i, KeySignatureToken):
                                    current_signature.last_key_signature_row = i
                                elif spine.isContentOfType(i, TimeSignatureToken):
                                    current_signature.last_time_signature_row = i
                                elif spine.isContentOfType(i, MeterSymbolToken):
                                    current_signature.last_meter_symbol_row = i

                        row_result += content
            if not empty:
                row_contents.append(row_result)
            else:
                row_contents.append(None)  # in order to maintain the indexes

            signatures_at_each_row.append(current_signature)

        # if last_header_row is None:
        #     raise Exception('No header row found')
        #
        # if last_clef_row is None:
        #     raise Exception('No clef row found')
        #
        # if last_time_signature_row is None and last_meter_symbol_row is None:
        #     raise Exception('No time signature or meter symbol row found')

        result = ''
        if options.from_measure is None and options.to_measure is None:
            for row_content in row_contents:
                if row_content:
                    result += row_content
                    result += '\n'
        else:
            if options.from_measure:
                self.checkMeasure(options.from_measure)
            else:
                options.from_measure = 0

            if options.to_measure:
                self.checkMeasure(options.to_measure)
            else:
                options.to_measure = len(self.measure_start_rows)

            from_row = self.measure_start_rows[options.from_measure - 1] - 1  # measures and rows are counted from 1
            if options.to_measure == len(self.measure_start_rows):
                to_row = self.measure_start_rows[options.to_measure - 1]
            else:
                to_row = self.measure_start_rows[options.to_measure]  # to the next one
            signature = signatures_at_each_row[from_row]

            # first, attach the signatures if not in the exported range
            result = self.addSignatureRowIfRequired(row_contents, result, from_row, signature.last_header_row)
            result = self.addSignatureRowIfRequired(row_contents, result, from_row, signature.last_clef_row)
            result = self.addSignatureRowIfRequired(row_contents, result, from_row, signature.last_key_signature_row)
            result = self.addSignatureRowIfRequired(row_contents, result, from_row, signature.last_time_signature_row)
            result = self.addSignatureRowIfRequired(row_contents, result, from_row, signature.last_meter_symbol_row)

            for row in range(from_row, to_row):
                row_content = row_contents[row]
                if row_content:
                    result += row_content
                    result += '\n'

            if to_row < max_rows:
                row_content = ''
                for spine in self.spines:
                    if spine.spine_type in options.spine_types and not spine.isTerminated():
                        if len(row_content) > 0:
                            row_content += '\t'
                        row_content += '*-'
                result += row_content
                result += '\n'
        return result

    def addSignatureRowIfRequired(self, row_contents, result, from_row, signature_row):
        if signature_row is not None and signature_row < from_row:
            srow = row_contents[signature_row]
            result += srow
            result += '\n'
        return result

    def getErrorMessages(self):
        result = ''
        for err in self.errors:
            result += str(err)
            result += '\n'
        return result

    def hasErrors(self):
        return len(self.errors) > 0

    def has_token(self, token_goal: str):
        """
        Check if the importer has a specific token.

        Args:
            token_goal: The token to check.

        Returns:
            True if the importer has the token, False otherwise.

        Example:
            # Create the importer
            >>> hi = HumdrumImporter()

            # Read the file
            >>> hi.import_file('file.krn')

            # Check if the importer has a specific token
            >>> has_f_4_clef = hi.has_token('*clefF4')
        """
        for spine in self.spines:
            for row in spine.rows:
                if any(token.encoding == token_goal for token in row):
                    return True

        return False

    def has_category(self, token_category_goal: TokenCategory):
        """
        Check if the importer has a specific token.

        Args:
            token_category_goal: The token category to check.
        Returns:
            True if the importer has the token category, False otherwise.

        Example:
            # Create the importer
            >>> hi = HumdrumImporter()

            # Read the file
            >>> hi.import_file('file.krn')

            # Check if the importer has a specific token
            >>> has_barlines = hi.has_category(TokenCategory.BARLINES)

        """
        for spine in self.spines:
            for row in spine.rows:
                for token in row:
                    if token.category == token_category_goal:
                        return True
        return False

    def get_all_tokens(self, apply_strip: bool = True, remove_measure_numbers: bool = False, filter_by_categories: Iterable = None) -> list:
        """
        Get all the tokens in the importer.

        Args:
            apply_strip: If True, the tokens will be stripped. False otherwise. Default is True.
            remove_measure_numbers: If True, the measure numbers will be removed. False otherwise. Default is False.
            filter_by_categories: An Iterable (like a list) with the categories to filter the tokens. Default is None.\
                Only the tokens with the categories in the list will be returned.


        Returns:
            A list with all the tokens in the importer.

        Example:
            # Create the importer
            >>> hi = HumdrumImporter()

            # Read the file
            >>> hi.import_file('file.krn')

            # Get all the tokens
            >>> all_tokens = hi.get_all_tokens()

            # Get all the tokens without measure numbers
            >>> all_tokens = hi.get_all_tokens(remove_measure_numbers=True)

            # Get all the tokens without measure numbers and filtered by categories
            >>> all_tokens = hi.get_all_tokens(remove_measure_numbers=True, filter_by_categories=[TokenCategory.BARLINES, TokenCategory.FINGERING, TokenCategory.CORE])

            # Get all tokens used in the bekern codification
            >>> all_tokens = hi.get_all_tokens(remove_measure_numbers=True, filter_by_categories=BEKERN_CATEGORIES)

        """
        MEASURE_START = '='
        DIGITS_TO_REMOVE = string.digits
        result = []
        for spine in self.spines:
            for row in spine.rows:
                for token in row:
                    if filter_by_categories is not None and token.category not in filter_by_categories:
                        continue

                    if remove_measure_numbers and token.encoding.startswith(MEASURE_START):
                        token.encoding = token.encoding.lstrip(DIGITS_TO_REMOVE)

                    if apply_strip:
                        token.encoding = token.encoding.strip()

                    result.append(token.encoding)

        return result

    def get_unique_tokens(self, apply_strip: bool = True, remove_measure_numbers: bool = False, filter_by_categories: Iterable = None) -> list:
        """
        Get the unique tokens in the importer.

        Args:
            apply_strip: If True, the tokens will be stripped. False otherwise. Default is True.
            remove_measure_numbers: If True, the measure numbers will be removed. False otherwise. Default is False.
            filter_by_categories: An Iterable (like a list) with the categories to filter the tokens. Default is None.\
                Only the tokens with the categories in the list will be returned.

        Returns:
            A list with the unique tokens in the importer.

        Example:
            # Create the importer
            >>> hi = HumdrumImporter()

            # Read the file
            >>> hi.import_file('file.krn')

            # Get the unique tokens
            >>> unique_tokens = hi.get_unique_tokens()

            # Get the unique tokens without measure numbers
            >>> unique_tokens = hi.get_unique_tokens(remove_measure_numbers=True)

            # Get the unique tokens without measure numbers and filtered by categories
            >>> unique_tokens = hi.get_unique_tokens(remove_measure_numbers=True, filter_by_categories=[TokenCategory.BARLINES, TokenCategory.KEYSIGNATURE, TokenCategory.CORE])

            # Get the unique tokens used in the bekern codification
            >>> unique_tokens = hi.get_all_tokens(remove_measure_numbers=True, filter_by_categories=BEKERN_CATEGORIES)

        """
        all_tokens = self.get_all_tokens(apply_strip=apply_strip, remove_measure_numbers=remove_measure_numbers, filter_by_categories=filter_by_categories)
        return list(set(all_tokens))

    def is_voice_in_tessitura(self, voice: int, tessitura: tuple) -> bool:
        """
        Check if a voice is in a tessitura.

        Args:
            voice: The voice to check.
            tessitura: A tuple with the tessitura. The first element is the lower limit, and the second element is the upper limit.

        Returns:
            True if the voice is in the tessitura, False otherwise.

        Example:
            # Create the importer
            >>> hi = HumdrumImporter()

            # Read the file
            >>> hi.import_file('file.krn')

            # Check if the voice 1 is in the tessitura (C4, G4)
            >>> is_in_tessitura = hi.is_voice_in_tessitura(1, ('c4', 'g4'))
        """
        raise NotImplementedError('This method is not implemented yet.')   # TODO: Implementar el método
        min_tessitura = tessitura[0].lower()
        max_tessitura = tessitura[1].lower()

        all_tokens = None
        for row in self.spines[voice].rows:
            all_tokens = [token.encoding.lower() for token in row if isinstance(token.category, SpineOperationToken)]   # TODO: Buscar la categoria que solo deje pasar notas

        for token in all_tokens:
            if token < min_tessitura or token > max_tessitura:
                return False

        return True

    def __len__(self):
        """
        Get the number of spines in the importer.
        """
        return len(self.spines)

def get_kern_from_ekern(ekern_content: string) -> string:
    """
    Read the content of a **ekern file and return the **kern content.

    Args:
        ekern_content: The content of the **ekern file.
    Returns:
        The content of the **kern file.

    Example:
        ```python
        # Read **ekern file
        ekern_file = 'path/to/file.ekrn'
        with open(ekern_file, 'r') as file:
            ekern_content = file.read()

        # Get **kern content
        kern_content = get_kern_from_ekern(ekern_content)
        with open('path/to/file.krn', 'w') as file:
            file.write(kern_content)

        ```
    """
    content = ekern_content.replace("**ekern", "**kern")  # TODO Constante según las cabeceras
    content = content.replace(TOKEN_SEPARATOR, "")
    content = content.replace(DECORATION_SEPARATOR, "")

    return content


def ekern_to_krn(input_file, output_file) -> None:
    """
    Convert one .ekrn file to .krn file.

    Args:
        input_file: Filepath to the input **ekern
        output_file: Filepath to the output **kern
    Returns:
        None

    Example:
        # Convert .ekrn to .krn
        >>> ekern_to_krn('path/to/file.ekrn', 'path/to/file.krn')

        # Convert a list of .ekrn files to .krn files
        ```python
        ekrn_files = your_modue.get_files()

        # Use the wrapper to avoid stopping the process if an error occurs
        def ekern_to_krn_wrapper(ekern_file, kern_file):
            try:
                ekern_to_krn(ekrn_files, output_folder)
            except Exception as e:
                print(f'Error:{e}')

        # Convert all the files
        for ekern_file in ekrn_files:
            output_file = ekern_file.replace('.ekrn', '.krn')
            ekern_to_krn_wrapper(ekern_file, output_file)
        ```
    """
    with open(input_file, 'r') as file:
        content = file.read()

    kern_content = get_kern_from_ekern(content)

    with open(output_file, 'w') as file:
        file.write(kern_content)


def kern_to_ekern(input_file, output_file) -> None:
    """
    Convert one .krn file to .ekrn file

    Args:
        input_file: Filepath to the input **kern
        output_file: Filepath to the output **ekern

    Returns:
        None

    Example:
        # Convert .krn to .ekrn
        >>> kern_to_ekern('path/to/file.krn', 'path/to/file.ekrn')

        # Convert a list of .krn files to .ekrn files
        ```python
        krn_files = your_module.get_files()

        # Use the wrapper to avoid stopping the process if an error occurs
        def kern_to_ekern_wrapper(krn_file, ekern_file):
            try:
                kern_to_ekern(krn_file, ekern_file)
            except Exception as e:
                print(f'Error:{e}')

        # Convert all the files
        for krn_file in krn_files:
            output_file = krn_file.replace('.krn', '.ekrn')
            kern_to_ekern_wrapper(krn_file, output_file)
        ```

    """
    importer = HumdrumImporter()
    importer.doImportFile(input_file)

    if len(importer.errors):
        raise Exception(f'ERROR: {input_file} has errors {importer.getErrorMessages()}')

    export_options = ExportOptions(spine_types=['**kern'], token_categories=BEKERN_CATEGORIES, kern_type=Encoding.eKern)
    exported_ekern = importer.doExport(export_options)

    with open(output_file, 'w') as file:
        file.write(exported_ekern)



from __future__ import annotations

from copy import copy, deepcopy
from collections import deque, defaultdict
from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Optional, Dict, Union
from collections.abc import Sequence
from queue import Queue

from kernpy.core import TokenCategory, CORE_HEADERS, TERMINATOR
from kernpy.core import MetacommentToken, AbstractToken, HeaderToken
from .transposer import transpose, Direction, NotationEncoding, AVAILABLE_INTERVALS
from .tokens import NoteRestToken, Subtoken
from .transposer import IntervalsByName


class SignatureNodes:
    """
    SignatureNodes class.

    This class is used to store the last signature nodes of a tree.
    It is used to keep track of the last signature nodes.

    Attributes: nodes (dict): A dictionary that stores the last signature nodes. This way, we can add several tokens
    without repetitions. - The key is the signature descendant token class (KeyToken, MeterSymbolToken, etc...) - The
    value = node

    """

    def __init__(self):
        """
        Create an instance of SignatureNodes. Initialize the nodes as an empty dictionary.

        Examples:
            >>> signature_nodes = SignatureNodes()
            >>> signature_nodes.nodes
            {}
        """
        self.nodes = {}

    def clone(self):
        """
        Create a deep copy of the SignatureNodes instance.
        Returns: A new instance of SignatureNodes with nodes copied.

        # TODO: This method is equivalent to the following code:
        # from copy import deepcopy
        # signature_nodes_to_copy = SignatureNodes()
        # ...
        # result = deepcopy(signature_nodes_to_copy)
        # It should be tested.
        """
        result = SignatureNodes()
        result.nodes = copy(self.nodes)
        return result

    def update(self, node):
        self.nodes[node.token.__class__.__name__] = node


class TreeTraversalInterface(ABC):
    """
    TreeTraversalInterface class.

    This class is used to traverse the tree. The `TreeTraversalInterface` class is responsible for implementing
    the `visit` method.
    """

    @abstractmethod
    def visit(self, node):
        pass


class Node:
    """
    Node class.

    This class represents a node in a tree.
    The `Node` class is responsible for storing the main information of the **kern file.

    Attributes:
        id(int): The unique id of the node.
        token(Optional[AbstractToken]): The specific token of the node. The token can be a `KeyToken`, `MeterSymbolToken`, etc...
        parent(Optional['Node']): A reference to the parent `Node`. If the parent is the root, the parent is None.
        children(List['Node']): A list of the children `Node`.
        stage(int): The stage of the node in the tree. The stage is similar to a row in the **kern file.
        last_spine_operator_node(Optional['Node']): The last spine operator node.
        last_signature_nodes(Optional[SignatureNodes]): A reference to the last `SignatureNodes` instance.
        header_node(Optional['Node']): The header node.
    """
    NextID = 1  # static counter

    def __init__(self,
                 stage: int,
                 token: Optional[AbstractToken],
                 parent: Optional['Node'],
                 last_spine_operator_node: Optional['Node'],
                 last_signature_nodes: Optional[SignatureNodes],
                 header_node: Optional['Node']
                 ):
        """
        Create an instance of Node.

        Args:
            stage (int): The stage of the node in the tree. The stage is similar to a row in the **kern file.
            token (Optional[AbstractToken]): The specific token of the node. The token can be a `KeyToken`, `MeterSymbolToken`, etc...
            parent (Optional['Node']): A reference to the parent `Node`. If the parent is the root, the parent is None.
            last_spine_operator_node (Optional['Node']): The last spine operator node.
            last_signature_nodes (Optional[SignatureNodes]): A reference to the last `SignatureNodes` instance.
            header_node (Optional['Node']): The header node.
        """
        self.id = Node.NextID
        Node.NextID += 1
        self.token = token
        self.parent = parent
        self.children = []
        self.stage = stage
        self.header_node = header_node
        if last_signature_nodes is not None:
            self.last_signature_nodes = last_signature_nodes.clone()  #TODO Documentar todo esto - composición
            # self.last_signature_nodes = copy.deepcopy(last_signature_nodes) # TODO: Ver en SignatureNodes.clone
        else:
            self.last_signature_nodes = SignatureNodes()
        self.last_spine_operator_node = last_spine_operator_node

    def count_nodes_by_stage(self) -> List[int]:
        """
        Count the number of nodes in each stage of the tree.

        Examples:
            >>> node = Node(0, None, None, None, None, None)
            >>> ...
            >>> node.count_nodes_by_stage()
            [2, 2, 2, 2, 3, 3, 3, 2]

        Returns:
            List[int]: A list with the number of nodes in each stage of the tree.
        """
        level_counts = defaultdict(int)
        queue = deque([(self, 0)])  # (node, level)
        # breadth-first search (BFS)
        while queue:
            node, level = queue.popleft()
            level_counts[level] += 1
            for child in node.children:
                queue.append((child, level + 1))

        # Convert the level_counts dictionary to a list of counts
        max_level = max(level_counts.keys())
        counts_by_level = [level_counts[level] for level in range(max_level + 1)]

        return counts_by_level

    def dfs(self, tree_traversal: TreeTraversalInterface):
        """
        Depth-first search (DFS)

        Args:
            tree_traversal (TreeTraversalInterface): The tree traversal interface. Object used to visit the nodes of the tree.
        """
        node = self
        tree_traversal.visit(node)
        for child in self.children:
            child.dfs(tree_traversal)

    def dfs_iterative(self, tree_traversal: TreeTraversalInterface):
        """
        Depth-first search (DFS). Iterative version.

        Args:
            tree_traversal (TreeTraversalInterface): The tree traversal interface. Object used to visit the nodes of the tree.

        Returns: None
        """
        stack = [self]
        while stack:
            node = stack.pop()
            tree_traversal.visit(node)
            stack.extend(reversed(node.children))  # Add children in reverse order to maintain DFS order

    def __eq__(self, other):
        """
        Compare two nodes.

        Args:
            other: The other node to compare.

        Returns: True if the nodes are equal, False otherwise.
        """
        if other is None or not isinstance(other, Node):
            return False

        return self.id == other.id

    def __ne__(self, other):
        """
        Compare two nodes.

        Args:
            other: The other node to compare.

        Returns: True if the nodes are not equal, False otherwise.
        """
        return not self.__eq__(other)

    def __hash__(self):
        """
        Get the hash of the node.

        Returns: The hash of the node.
        """
        return hash(self.id)

    def __str__(self):
        """
        Get the string representation of the node.

        Returns: The string representation of the node.
        """
        return f"{{{self.stage}: {self.token}}}"


class BoundingBoxMeasures:
    """
    BoundingBoxMeasures class.
    """

    def __init__(
            self,
            bounding_box,
            from_measure: int,
            to_measure: int
    ):
        """
        Create an instance of BoundingBoxMeasures.

        Args:
            bounding_box: The bounding box object of the node.
            from_measure (int): The first measure of the score in the BoundingBoxMeasures object.
            to_measure (int): The last measure of the score in the BoundingBoxMeasures object.
        """
        self.from_measure = from_measure
        self.to_measure = to_measure
        self.bounding_box = bounding_box


class MultistageTree:
    """
    MultistageTree class.
    """

    def __init__(self):
        """
        Constructor for MultistageTree class.

        Create an empty Node object to serve as the root, \
        and start the stages list by placing this root node inside a new list.

        """
        self.root = Node(0, None, None, None, None, None)
        self.stages = []  # First stage (0-index) is the root (Node with None token and header_node). The core header is in stage 1.
        self.stages.append([self.root])

    def add_node(
            self,
            stage: int,
            parent: Node,
            token: Optional[AbstractToken],
            last_spine_operator_node: Optional[Node],
            previous_signature_nodes: Optional[SignatureNodes],
            header_node: Optional[Node] = None
    ) -> Node:
        """
        Add a new node to the tree.
        Args:
            stage (int):
            parent (Node):
            token (Optional[AbstractToken]):
            last_spine_operator_node (Optional[Node]):
            previous_signature_nodes (Optional[SignatureNodes]):
            header_node (Optional[Node]):

        Returns: Node - The added node object.

        """
        node = Node(stage, token, parent, last_spine_operator_node, previous_signature_nodes, header_node)
        if stage == len(self.stages):
            self.stages.append([node])
        elif stage > len(self.stages):
            raise ValueError(f'Cannot add node in stage {stage} when there are only {len(self.stages)} stages')
        else:
            self.stages[stage].append(node)

        parent.children.append(node)
        return node

    def dfs(self, visit_method) -> None:
        """
        Depth-first search (DFS)

        Args:
            visit_method (TreeTraversalInterface): The tree traversal interface.

        Returns: None

        """
        self.root.dfs(visit_method)

    def dfs_iterative(self, visit_method) -> None:
        """
        Depth-first search (DFS). Iterative version.

        Args:
            visit_method (TreeTraversalInterface): The tree traversal interface.

        Returns: None

        """
        self.root.dfs_iterative(visit_method)

    def __deepcopy__(self, memo):
        """
        Create a deep copy of the MultistageTree object.
        """
        # Create a new empty MultistageTree object
        new_tree = MultistageTree()

        # Deepcopy the root
        new_tree.root = deepcopy(self.root, memo)

        # Deepcopy the stages list
        new_tree.stages = deepcopy(self.stages, memo)

        return new_tree


class Document:
    """
    Document class.

    This class store the score content using an agnostic tree structure.

    Attributes:
        tree (MultistageTree): The tree structure of the document where all the nodes are stored. \
            Each stage of the tree corresponds to a row in the Humdrum **kern file encoding.
        measure_start_tree_stages (List[List[Node]]): The list of nodes that corresponds to the measures. \
            Empty list by default.
            The index of the list is starting from 1. Rows after removing empty lines and line comments
        page_bounding_boxes (Dict[int, BoundingBoxMeasures]): The dictionary of page bounding boxes. \
            - key: page number
            - value: BoundingBoxMeasures object
        header_stage (int): The index of the stage that contains the headers. None by default.
    """

    def __init__(self, tree: MultistageTree):
        """
        Constructor for Document class.

        Args:
            tree (MultistageTree): The tree structure of the document where all the nodes are stored.
        """
        self.tree = tree  # TODO: ? Should we use copy.deepcopy() here?
        self.measure_start_tree_stages = []
        self.page_bounding_boxes = {}
        self.header_stage = None

    FIRST_MEASURE = 1

    def get_header_stage(self) -> Union[List[Node], List[List[Node]]]:
        """
        Get the Node list of the header stage.

        Returns: (Union[List[Node], List[List[Node]]]) The Node list of the header stage.

        Raises: Exception - If the document has no header stage.
        """
        if self.header_stage:
            return self.tree.stages[self.header_stage]
        else:
            raise Exception('No header stage found')

    def get_leaves(self) -> List[Node]:
        """
        Get the leaves of the tree.

        Returns: (List[Node]) The leaves of the tree.
        """
        return self.tree.stages[len(self.tree.stages) - 1]

    def get_spine_count(self) -> int:
        """
        Get the number of spines in the document.

        Returns (int): The number of spines in the document.
        """
        return len(self.get_header_stage())  # TODO: test refactor

    def get_first_measure(self) -> int:
        """
        Get the index of the first measure of the document.

        Returns: (Int) The index of the first measure of the document.

        Raises: Exception - If the document has no measures.

        Examples:
            >>> import kernpy as kp
            >>> document, err = kp.read('score.krn')
            >>> document.get_first_measure()
            1
        """
        if len(self.measure_start_tree_stages) == 0:
            raise Exception('No measures found')

        return self.FIRST_MEASURE

    def measures_count(self) -> int:
        """
        Get the index of the last measure of the document.

        Returns: (Int) The index of the last measure of the document.

        Raises: Exception - If the document has no measures.

        Examples:
            >>> document, _ = kernpy.read('score.krn')
            >>> document.measures_count()
            10
            >>> for i in range(document.get_first_measure(), document.measures_count() + 1):
            >>>   options = kernpy.ExportOptions(from_measure=i, to_measure=i+4)
        """
        if len(self.measure_start_tree_stages) == 0:
            raise Exception('No measures found')

        return len(self.measure_start_tree_stages)

    def get_metacomments(self, KeyComment: Optional[str] = None, clear: bool = False) -> List[str]:
        """
        Get all metacomments in the document

        Args:
            KeyComment: Filter by a specific metacomment key: e.g. Use 'COM' to get only comments starting with\
                '!!!COM: '. If None, all metacomments are returned.
            clear: If True, the metacomment key is removed from the comment. E.g. '!!!COM: Coltrane' -> 'Coltrane'.\
                If False, the metacomment key is kept. E.g. '!!!COM: Coltrane' -> '!!!COM: Coltrane'. \
                The clear functionality is equivalent to the following code:
                ```python
                comment = '!!!COM: Coltrane'
                clean_comment = comment.replace(f"!!!{KeyComment}: ", "")
                ```
                Other formats are not supported.

        Returns: A list of metacomments.

        Examples:
            >>> document.get_metacomments()
            ['!!!COM: Coltrane', '!!!voices: 1', '!!!OPR: Blue Train']
            >>> document.get_metacomments(KeyComment='COM')
            ['!!!COM: Coltrane']
            >>> document.get_metacomments(KeyComment='COM', clear=True)
            ['Coltrane']
            >>> document.get_metacomments(KeyComment='non_existing_key')
            []
        """
        traversal = MetacommentsTraversal()
        self.tree.dfs_iterative(traversal)
        result = []
        for metacomment in traversal.metacomments:
            if KeyComment is None or metacomment.encoding.startswith(f"!!!{KeyComment}"):
                new_comment = metacomment.encoding
                if clear:
                    new_comment = metacomment.encoding.replace(f"!!!{KeyComment}: ", "")
                result.append(new_comment)

        return result

    @classmethod
    def tokens_to_encodings(cls, tokens: Sequence[AbstractToken]):
        """
        Get the encodings of a list of tokens.

        The method is equivalent to the following code:
            >>> tokens = kp.get_all_tokens()
            >>> [token.encoding for token in tokens if token.encoding is not None]

        Args:
            tokens (Sequence[AbstractToken]): list - A list of tokens.

        Returns: List[str] - A list of token encodings.

        Examples:
            >>> tokens = document.get_all_tokens()
            >>> Document.tokens_to_encodings(tokens)
            ['!!!COM: Coltrane', '!!!voices: 1', '!!!OPR: Blue Train']
        """
        encodings = [token.encoding for token in tokens if token.encoding is not None]
        return encodings

    def get_all_tokens(self, filter_by_categories: Optional[Sequence[TokenCategory]] = None) -> List[AbstractToken]:
        """
        Args:
            filter_by_categories (Optional[Sequence[TokenCategory]]): A list of categories to filter the tokens. If None, all tokens are returned.

        Returns:
            List[AbstractToken] - A list of all tokens.

        Examples:
            >>> tokens = document.get_all_tokens()
            >>> Document.tokens_to_encodings(tokens)
            >>> [type(t) for t in tokens]
            [<class 'kernpy.core.token.Token'>, <class 'kernpy.core.token.Token'>, <class 'kernpy.core.token.Token'>]
        """
        computed_categories = TokenCategory.valid(include=filter_by_categories)
        traversal = TokensTraversal(False, computed_categories)
        self.tree.dfs_iterative(traversal)
        return traversal.tokens

    def get_all_tokens_encodings(
            self,
            filter_by_categories: Optional[Sequence[TokenCategory]] = None
    ) -> List[str]:
        """
        Args:
            filter_by_categories (Optional[Sequence[TokenCategory]]): A list of categories to filter the tokens. If None, all tokens are returned.


        Returns:
            list[str] - A list of all token encodings.

        Examples:
            >>> tokens = document.get_all_tokens_encodings()
            >>> Document.tokens_to_encodings(tokens)
            ['!!!COM: Coltrane', '!!!voices: 1', '!!!OPR: Blue Train']
        """
        tokens = self.get_all_tokens(filter_by_categories)
        return Document.tokens_to_encodings(tokens)

    def get_unique_tokens(
            self,
            filter_by_categories: Optional[Sequence[TokenCategory]] = None
    ) -> List[AbstractToken]:
        """
        Get unique tokens.

        Args:
            filter_by_categories (Optional[Sequence[TokenCategory]]): A list of categories to filter the tokens. If None, all tokens are returned.

        Returns:
            List[AbstractToken] - A list of unique tokens.

        """
        computed_categories = TokenCategory.valid(include=filter_by_categories)
        traversal = TokensTraversal(True, computed_categories)
        self.tree.dfs_iterative(traversal)
        return traversal.tokens

    def get_unique_token_encodings(
            self,
            filter_by_categories: Optional[Sequence[TokenCategory]] = None
    ) -> List[str]:
        """
        Get unique token encodings.

        Args:
            filter_by_categories (Optional[Sequence[TokenCategory]]): A list of categories to filter the tokens. If None, all tokens are returned.

        Returns: List[str] - A list of unique token encodings.

        """
        tokens = self.get_unique_tokens(filter_by_categories)
        return Document.tokens_to_encodings(tokens)

    def get_voices(self, clean: bool = False):
        """
        Get the voices of the document.

        Args
            clean (bool): Remove the first '!' from the voice name.

        Returns: A list of voices.

        Examples:
            >>> document.get_voices()
            ['!sax', '!piano', '!bass']
            >>> document.get_voices(clean=True)
            ['sax', 'piano', 'bass']
            >>> document.get_voices(clean=False)
            ['!sax', '!piano', '!bass']
        """
        from kernpy.core import TokenCategory
        voices = self.get_all_tokens(filter_by_categories=[TokenCategory.INSTRUMENTS])

        if clean:
            voices = [voice[1:] for voice in voices]
        return voices

    def clone(self):
        """
        Create a deep copy of the Document instance.

        Returns: A new instance of Document with the tree copied.

        """
        result = Document(copy(self.tree))
        result.measure_start_tree_stages = copy(self.measure_start_tree_stages)
        result.page_bounding_boxes = copy(self.page_bounding_boxes)
        result.header_stage = copy(self.header_stage)

        return result

    def append_spines(self, spines) -> None:
        """
        Append the spines directly to current document tree.

        Args:
            spines(list): A list of spines to append.

        Returns: None

        Examples:
            >>> import kernpy as kp
            >>> doc, _ = kp.read('score.krn')
            >>> spines = [
            >>> '4e\t4f\t4g\t4a\n4b\t4c\t4d\t4e\n=\t=\t=\t=\n',
            >>> '4c\t4d\t4e\t4f\n4g\t4a\t4b\t4c\n=\t=\t=\t=\n',
           >>> ]
           >>> doc.append_spines(spines)
           None
        """
        raise NotImplementedError()
        if len(spines) != self.get_spine_count():
            raise Exception(f"Spines count mismatch: {len(spines)} != {self.get_spine_count()}")

        for spine in spines:
            return

    def add(self, other: 'Document', *, check_core_spines_only: Optional[bool] = False) -> 'Document':
        """
        Concatenate one document to the current document: Modify the current object!

        Args:
            other: The document to concatenate.
            check_core_spines_only: If True, only the core spines (**kern and **mens) are checked. If False, all spines are checked.

        Returns ('Document'): The current document (self) with the other document concatenated.
        """
        if not Document.match(self, other, check_core_spines_only=check_core_spines_only):
            raise Exception(f'Documents are not compatible for addition. '
                            f'Headers do not match with check_core_spines_only={check_core_spines_only}. '
                            f'self: {self.get_header_nodes()}, other: {other.get_header_nodes()}. ')

        current_header_nodes = self.get_header_stage()
        other_header_nodes = other.get_header_stage()

        current_leaf_nodes = self.get_leaves()
        flatten = lambda lst: [item for sublist in lst for item in sublist]
        other_first_level_children = [flatten(c.children) for c in other_header_nodes]  # avoid header stage

        for current_leaf, other_first_level_child in zip(current_leaf_nodes, other_first_level_children, strict=False):
            # Ignore extra spines from other document.
            # But if there are extra spines in the current document, it will raise an exception.
            if current_leaf.token.encoding == TERMINATOR:
                # remove the '*-' token from the current document
                current_leaf_index = current_leaf.parent.children.index(current_leaf)
                current_leaf.parent.children.pop(current_leaf_index)
                current_leaf.parent.children.insert(current_leaf_index, other_first_level_child)

            self.tree.add_node(
                stage=len(self.tree.stages) - 1,  # TODO: check offset 0, +1, -1 ????
                parent=current_leaf,
                token=other_first_level_child.token,
                last_spine_operator_node=other_first_level_child.last_spine_operator_node,
                previous_signature_nodes=other_first_level_child.last_signature_nodes,
                header_node=other_first_level_child.header_node
            )

        return self

    def get_header_nodes(self) -> List[HeaderToken]:
        """
        Get the header nodes of the current document.

        Returns: List[HeaderToken]: A list with the header nodes of the current document.
        """
        return [token for token in self.get_all_tokens(filter_by_categories=None) if isinstance(token, HeaderToken)]

    def get_spine_ids(self) -> List[int]:
        """
                Get the indexes of the current document.

                Returns List[int]: A list with the indexes of the current document.

                Examples:
                    >>> document.get_all_spine_indexes()
                    [0, 1, 2, 3, 4]
                """
        header_nodes = self.get_header_nodes()
        return [node.spine_id for node in header_nodes]

    def frequencies(self, token_categories: Optional[Sequence[TokenCategory]] = None) -> Dict:
        """
        Frequency of tokens in the document.


        Args:
            token_categories (Optional[Sequence[TokenCategory]]): If None, all tokens are considered.
        Returns (Dict):
            A dictionary with the category and the number of occurrences of each token.

        """
        tokens = self.get_all_tokens(filter_by_categories=token_categories)
        frequencies = {}
        for t in tokens:
            if t.encoding in frequencies:
                frequencies[t.encoding]['occurrences'] += 1
            else:
                frequencies[t.encoding] = {
                    'occurrences': 1,
                    'category': t.category.name,
                }

        return frequencies

    def split(self) -> List['Document']:
        """
        Split the current document into a list of documents, one for each **kern spine.
        Each resulting document will contain one **kern spine along with all non-kern spines.

        Returns:
            List['Document']: A list of documents, where each document contains one **kern spine
            and all non-kern spines from the original document.

        Examples:
            >>> document.split()
            [<Document: score.krn>, <Document: score.krn>, <Document: score.krn>]
        """
        raise NotImplementedError
        new_documents = []
        self_document_copy = deepcopy(self)
        kern_header_nodes = [node for node in self_document_copy.get_header_nodes() if node.encoding == '**kern']
        other_header_nodes = [node for node in self_document_copy.get_header_nodes() if node.encoding != '**kern']
        spine_ids = self_document_copy.get_spine_ids()

        for header_node in kern_header_nodes:
            if header_node.spine_id not in spine_ids:
                continue

            spine_ids.remove(header_node.spine_id)

            new_tree = deepcopy(self.tree)
            prev_node = new_tree.root
            while not isinstance(prev_node, HeaderToken):
                prev_node = prev_node.children[0]

            if not prev_node or not isinstance(prev_node, HeaderToken):
                raise Exception(f'Header node not found: {prev_node} in {header_node}')

            new_children = list(filter(lambda x: x.spine_id == header_node.spine_id, prev_node.children))
            new_tree.root = new_children

            new_document = Document(new_tree)

            new_documents.append(new_document)

        return new_documents

    @classmethod
    def to_concat(cls, first_doc: 'Document', second_doc: 'Document', deep_copy: bool = True) -> 'Document':
        """
        Concatenate two documents.

        Args:
            first_doc (Document): The first document.
            second_doc (Document: The second document.
            deep_copy (bool): If True, the documents are deep copied. If False, the documents are shallow copied.

        Returns: A new instance of Document with the documents concatenated.
        """
        first_doc = first_doc.clone() if deep_copy else first_doc
        second_doc = second_doc.clone() if deep_copy else second_doc
        first_doc.add(second_doc)

        return first_doc

    @classmethod
    def match(cls, a: 'Document', b: 'Document', *, check_core_spines_only: Optional[bool] = False) -> bool:
        """
        Match two documents. Two documents match if they have the same spine structure.

        Args:
            a (Document): The first document.
            b (Document): The second document.
            check_core_spines_only (Optional[bool]): If True, only the core spines (**kern and **mens) are checked. If False, all spines are checked.

        Returns: True if the documents match, False otherwise.

        Examples:

        """
        if check_core_spines_only:
            return [token.encoding for token in a.get_header_nodes() if token.encoding in CORE_HEADERS] \
                == [token.encoding for token in b.get_header_nodes() if token.encoding in CORE_HEADERS]
        else:
            return [token.encoding for token in a.get_header_nodes()] \
                == [token.encoding for token in b.get_header_nodes()]


    def to_transposed(self, interval: str, direction: str = Direction.UP.value) -> 'Document':
        """
        Create a new document with the transposed notes without modifying the original document.

        Args:
            interval (str): The name of the interval to transpose. It can be 'P4', 'P5', 'M2', etc. Check the \
             kp.AVAILABLE_INTERVALS for the available intervals.
            direction (str): The direction to transpose. It can be 'up' or 'down'.

        Returns:

        """
        if interval not in AVAILABLE_INTERVALS:
            raise ValueError(
                f"Interval {interval!r} is not available. "
                f"Available intervals are: {AVAILABLE_INTERVALS}"
            )

        if direction not in (Direction.UP.value, Direction.DOWN.value):
            raise ValueError(
                f"Direction {direction!r} is not available. "
                f"Available directions are: "
                f"{Direction.UP.value!r}, {Direction.DOWN.value!r}"
            )

        new_document = self.clone()

        # BFS through the tree
        root = new_document.tree.root
        queue = Queue()
        queue.put(root)

        while not queue.empty():
            node = queue.get()

            if isinstance(node.token, NoteRestToken):
                orig_token = node.token

                new_subtokens = []
                transposed_pitch_encoding = None

                # Transpose each pitch subtoken in the pitch–duration list
                for subtoken in orig_token.pitch_duration_subtokens:
                    if subtoken.category == TokenCategory.PITCH:
                        # transpose() returns a new pitch subtoken
                        tp = transpose(
                            input_encoding=subtoken.encoding,
                            interval=IntervalsByName[interval],
                            direction=direction,
                            input_format=NotationEncoding.HUMDRUM.value,
                            output_format=NotationEncoding.HUMDRUM.value,
                        )
                        new_subtokens.append(Subtoken(tp, subtoken.category))
                        transposed_pitch_encoding = tp
                    else:
                        # leave duration subtokens untouched
                        new_subtokens.append(Subtoken(subtoken.encoding, subtoken.category))

                # Replace the node’s token with a new NoteRestToken
                node.token = NoteRestToken(
                    encoding=transposed_pitch_encoding,
                    pitch_duration_subtokens=new_subtokens,
                    decoration_subtokens=orig_token.decoration_subtokens,
                )

            # enqueue children
            for child in node.children:
                queue.put(child)

        # Return the transposed clone
        return new_document


    def __iter__(self):
        """
        Get the indexes to export all the document.

        Returns: An iterator with the indexes to export the document.
        """
        return iter(range(self.get_first_measure(), self.measures_count() + 1))

    def __next__(self):
        """
        Get the next index to export the document.

        Returns: The next index to export the document.
        """
        return next(iter(range(self.get_first_measure(), self.measures_count() + 1)))


# tree traversal utils
class MetacommentsTraversal(TreeTraversalInterface):
    def __init__(self):
        self.metacomments = []

    def visit(self, node):
        if isinstance(node.token, MetacommentToken):
            self.metacomments.append(node.token)


class TokensTraversal(TreeTraversalInterface):
    def __init__(
            self,
            non_repeated: bool,
            filter_by_categories
    ):
        """
        Create an instance of `TokensTraversal`.
        Args:
            non_repeated: If True, only unique tokens are returned. If False, all tokens are returned.
            filter_by_categories: A list of categories to filter the tokens. If None, all tokens are returned.
        """
        self.tokens = []
        self.seen_encodings = []
        self.non_repeated = non_repeated
        self.filter_by_categories = [t for t in TokenCategory] if filter_by_categories is None else filter_by_categories

    def visit(self, node):
        if (node.token
                and (not self.non_repeated or node.token.encoding not in self.seen_encodings)
                and (self.filter_by_categories is None or node.token.category in self.filter_by_categories)
        ):
            self.tokens.append(node.token)
            if self.non_repeated:
                self.seen_encodings.append(node.token.encoding)


class TraversalFactory:
    class Categories(Enum):
        METACOMMENTS = "metacomments"
        TOKENS = "tokens"

    @classmethod
    def create(
            cls,
            traversal_type: str,
            non_repeated: bool,
            filter_by_categories: Optional[Sequence[TokenCategory]]
    ) -> TreeTraversalInterface:
        """
        Create an instance of `TreeTraversalInterface` based on the `traversal_type`.
        Args:
            non_repeated:
            filter_by_categories:
            traversal_type: The type of traversal to use. Possible values are:
                - "metacomments"
                - "tokens"

        Returns: An instance of `TreeTraversalInterface`.
        """
        if traversal_type == cls.Categories.METACOMMENTS.value:
            return MetacommentsTraversal()
        elif traversal_type == cls.Categories.TOKENS.value:
            return TokensTraversal(non_repeated, filter_by_categories)

        raise ValueError(f"Unknown traversal type: {traversal_type}")




import sys

import requests
from PIL import Image
from io import BytesIO
import os
import json
from typing import List, Optional

from kernpy import ExportOptions, BEKERN_CATEGORIES, Importer, Exporter, Document, Encoding, read


# This script creates the Polish dataset from the kern files.
# It downloads both the systems and full pages
DEFAULT_IIIF_ID = '/full/full/0/default.jpg'

LOG_FILENAME = 'polish_index.json'


def get_image_urls(_manifest_url):
    # It returns the URL of pages tagged with a page number
    # It returns a map with the page label as key, and the URL as
    response = requests.get(_manifest_url)
    manifest_data = response.json()
    result = {}

    # The corpus contains two kinds of IIIF manifests
    if 'sequences' in manifest_data:
        for sequence in manifest_data['sequences']:
            for canvas in sequence['canvases']:
                image_id = canvas['images'][0]['resource']['service']['@id']
                page_label = canvas['label'] if 'label' in canvas else None
                result[page_label] = image_id
    else:
        items = manifest_data.get('items', [])
        for item in items:
            pl = item.get('label').get('pl')
            if pl:
                page_label = pl[0]
                if page_label != '[]':
                    image_id = item.get('items')[0].get('items')[0].get('id', '')
                    if image_id.endswith(DEFAULT_IIIF_ID):
                        image_id = image_id[:-len(DEFAULT_IIIF_ID)]
                    result[page_label] = image_id

    # print(f'Items: ', len(items))

    return result


def download_and_save_image(url, save_path):
    try:
        # Make a GET request to the URL
        response = requests.get(url)
        response.raise_for_status()  # Raise an exception for bad responses

        # Open the image using Pillow
        image = Image.open(BytesIO(response.content))

        # Save the image to the specified path
        image.save(save_path, format='JPEG')

        print(f"Image downloaded and saved to: {save_path}")
    except Exception as e:
        print(f"An error occurred: {e}")


def factory_get_kern_type_exporter(kern_type: str) -> Encoding:
    """
    Factory method to get the Encoding

    Args:
        kern_type (str): the type of kern exporter. It can be 'krn' or 'ekrn'

    Returns:
        Encoding: the Encoding instance
    """
    if kern_type == 'krn':
        return Encoding.normalizedKern
    elif kern_type == 'ekrn':
        return Encoding.eKern
    else:
        raise Exception(f'Unknown export kern type: {kern_type}')


def extract_and_save_measures(document, from_measure, to_measure, krn_path, exporter_kern_type='ekrn'):
    exporter_kern_type = factory_get_kern_type_exporter(exporter_kern_type)
    export_options = ExportOptions(spine_types=['**kern'], token_categories=BEKERN_CATEGORIES, kern_type=exporter_kern_type)
    export_options.from_measure = from_measure
    export_options.to_measure = to_measure
    exporter = Exporter()
    exported_ekern = exporter.export_string(document, export_options)
    with open(krn_path, "w") as f:
        f.write(exported_ekern)


def download_and_save_page_images(document, _output_path, map_page_label_iiif_ids, page_bounding_boxes, log_filename, exporter_kern_type='ekrn'):
    print(f'Bounding boxes {page_bounding_boxes}')

    for page_label, bounding_box_measure in page_bounding_boxes.items():
        page_iiif_id = map_page_label_iiif_ids.get(page_label)
        if page_iiif_id is None and page_label.startswith('#'):  # sometimes it's wrongly tagged without the #
            page_iiif_id = map_page_label_iiif_ids.get(page_label[1:])

        if page_iiif_id is not None:
            bounding_box = bounding_box_measure.bounding_box
            print(f"Page: {page_label}, "
                  f"Bounding box: {bounding_box}, "
                  f"ID: {page_iiif_id}, "
                  f"from bar {bounding_box_measure.from_measure}, "
                  f"to bar {bounding_box_measure.to_measure}")
            url = os.path.join(page_iiif_id, bounding_box.xywh(), 'full', '0', 'default.jpg')
            print(url)
            image_path = os.path.join(_output_path, page_label + ".jpg")
            download_and_save_image(url, image_path)
            krn_path = os.path.join(_output_path, page_label + f'.{exporter_kern_type}')
            extract_and_save_measures(document, bounding_box_measure.from_measure, bounding_box_measure.to_measure - 1,
                                      krn_path, exporter_kern_type=exporter_kern_type)
            add_log(document, krn_path, log_filename=log_filename)
        else:
            raise Exception(f'Cannot find IIIF id for page with label "{page_label}"')


def findIIIFIds(document):
    iiifTag = "!!!IIIF:"
    for metacomment_token in document.get_metacomments():
        encoding = metacomment_token
        if encoding.startswith(iiifTag):
            url = encoding[len(iiifTag):].strip()
            print(f'Reading IIIF manifest from {url}')
            return get_image_urls(url)
    raise Exception('Cannot find any IIIF metacomment')


def is_valid_document(document, kern_spines_filter) -> bool:
    if kern_spines_filter is None:
        return True

    exporter = Exporter()
    kern_types = exporter.get_spine_types(document, spine_types=['**kern'])
    return len(kern_types) == int(kern_spines_filter)


def convert_and_download_file(input_kern, _output_path, log_filename, kern_spines_filter: int = None, exporter_kern_type='ekrn') -> None:
    document, errors = read(input_kern)
    if len(errors) > 0:
        print(f'ERRORS when kernpy.read:{input_kern} has errors {errors}\nContinue...', file=sys.stderr)
        raise Exception(f'ERRORS when kernpy.read: {input_kern}. Has errors: {errors}')

    if not is_valid_document(document, kern_spines_filter):
        return

    map_page_label_IIIF_ids = findIIIFIds(document)
    download_and_save_page_images(document, _output_path, map_page_label_IIIF_ids, document.page_bounding_boxes,
                                  log_filename=log_filename, exporter_kern_type=exporter_kern_type)


def search_files_with_string(root_folder, target_string):
    matching_files = []

    for foldername, subfolders, filenames in os.walk(root_folder):
        for filename in filenames:
            if filename.endswith('.krn'):
                file_path = os.path.join(foldername, filename)
                try:
                    with open(file_path, 'r', encoding='utf-8') as file:
                        content = file.read()
                        if target_string in content:
                            relative_path = os.path.relpath(file_path, root_folder)
                            matching_files.append(relative_path)
                except Exception as e:
                    print(f"Error reading {file_path}: {e}")

    return matching_files


def remove_extension(file_name):
    # Using os.path.splitext to split the file name and extension
    base_name, _ = os.path.splitext(file_name)
    return base_name


def add_log(document: Document, path, log_filename) -> None:
    try:
        def get_instruments(line):
            words = line.split(' ')
            instruments = []
            for i in range(len(words)):
                if words[i].isnumeric():
                    instruments.extend([words[i + 1]] * int(words[i]))
            return instruments

        def get_publish_date(line):
            if line is None or line == '':
                return 0

            clean_line = [char for char in line if char.isnumeric()]
            return int(''.join(clean_line))

        def round_publication_year(original_composer_date):
            try:
                if original_composer_date is None:
                    return 0
                start_date, end_date = original_composer_date.split('-')

                start_year = int(start_date.split('/')[0])
                end_year = int(end_date.split('/')[0])

                RATIO = 0.7  # date where the composer was most active
                return int(start_year + (end_year - start_year) * RATIO)
            except Exception as e:
                return -1

        def round_publication_year_v2(original_composer_date):
            def flatten(xss):
                return [x for xs in xss for x in xs]

            try:
                items_date = original_composer_date.split('/')
                clean_items = [item.replace(' ', '') for item in items_date]
                clean_items = [item.replace('~', '') for item in clean_items]
                split_again = [item.split('-') for item in clean_items]
                flatten_items = flatten(split_again)
                useful_items = [item for item in flatten_items if item.isnumeric()]
                year_items = [int(item) for item in useful_items if len(item) == 4]
                return int(year_items[0]) if len(year_items) > 0 else -3
            except Exception as e:
                return -2

        info = {
            'path': path,
            'publication_date': get_publish_date(document.get_metacomments('PDT')[0]) if document.get_metacomments(
                'PDT') else None,
            'original_publication_date_tag': True,
            'iiif': document.get_metacomments('IIIF')[0] if document.get_metacomments('IIIF') else None,
            'n_measures': len(document.tree.stages),
            'composer': document.get_metacomments('COM')[0] if document.get_metacomments('COM') else None,
            'composer_dates': document.get_metacomments('CDT')[0] if document.get_metacomments('CDT') else None,
            'tempo': document.get_metacomments('OTL')[0] if document.get_metacomments('OTL') else None,
            'piece_title': document.get_metacomments('OPR')[0] if document.get_metacomments('OPR') else None,
            'segment': document.get_metacomments('SEGMENT')[0] if document.get_metacomments('SEGMENT') else None,
            'n_voices': len(get_instruments(document.get_metacomments('AIN')[0])) if document.get_metacomments(
                'AIN') else 0,
            'instruments': get_instruments(document.get_metacomments('AIN')[0]) if document.get_metacomments(
                'AIN') else [],
            'unique_instruments': [
                *set(get_instruments(document.get_metacomments('AIN')[0]))] if document.get_metacomments('AIN') else [],
        }

        if info['publication_date'] in (0, 1, -1, -2) or info['publication_date'] is None:
            info['publication_date'] = round_publication_year(info['composer_dates'])
            info['original_publication_date_tag'] = False

        if info['publication_date'] in (0, 1, -1, -2) or info['publication_date'] is None:
            info['publication_date'] = round_publication_year_v2(info['composer_dates'])
            info['original_publication_date_tag'] = False

        with open(log_filename, 'a') as f:
            json.dump(info, f)
            f.write('\n')
    except Exception as e:
        print(f"Error adding log:{path}:{e}")


def remove_empty_dirs(directory):
    for root, dirs, files in os.walk(directory):
        for dir in dirs:
            full_dir = os.path.join(root, dir)
            if not os.listdir(full_dir):
                os.rmdir(full_dir)


def store_error_log(filename, msg: dict):
    with open(filename, encoding='utf-8', mode='a') as f:
        f.write(f'{json.dumps(msg)}\n')


def main(
        input_directory: str,
        output_directory: str,
        remove_empty_directories: Optional[bool] = True,
        kern_spines_filter: Optional[int] = 2,
        exporter_kern_type: Optional[str] = 'ekern'
) -> None:
    """
    Process the files in the input_directory and save the results in the output_directory.
    http requests are made to download the images.

    Args:
        input_directory (str): directory where the input files are found
        output_directory (str): directory where the output files are saved
        remove_empty_directories (Optional[bool]): remove empty directories when finish processing the files
        kern_spines_filter (Optional[int]): Only process files with the number of **kern spines specified.\
            Use it to export 2-voice files. Default is 2.\
            Use None to process all files.
        exporter_kern_type (Optional[str]): the type of kern exporter. It can be 'krn' or 'ekrn'



    Returns:
        None

    Examples:
        >>> main('/kern_files', '/output_ekern')
        None

        >>> main('/kern_files', '/output_ekern', remove_empty_directories=False)
        None

        >>> main('/kern_files', '/output_ekern', kern_spines_filter=2, remove_empty_directories=False)
        None

        >>> main('/kern_files', '/output_ekern', kern_spines_filter=None, remove_empty_directories=False)
        None

        >>> main('/kern_files', '/output_ekern', exporter_kern_type='krn', remove_empty_directories=True)
        None

        >>> main('/kern_files', '/output_ekern', exporter_kern_type='ekrn', remove_empty_directories=True, kern_spines_filter=2)
        None

    """
    print(f'Processing files in {input_directory} and saving to {output_directory}')
    kern_with_bboxes = search_files_with_string(input_directory, 'xywh')
    ok_files = []
    ko_files = []
    log_file = os.path.join(output_directory, LOG_FILENAME)
    print(f"{25*'='}"
          f"\nProcessing {len(kern_with_bboxes)} files."
          f"\nLog will be saved in {log_file}."
          f"\n{25*'='}")
    for kern in kern_with_bboxes:
        try:
            filename = remove_extension(kern)
            kern_path = os.path.join(input_directory, kern)
            output_kern_path = os.path.join(output_directory, filename)
            if not os.path.exists(output_kern_path):
                os.makedirs(output_kern_path)
            convert_and_download_file(kern_path, output_kern_path, log_filename=log_file, kern_spines_filter=kern_spines_filter, exporter_kern_type=exporter_kern_type)
            ok_files.append(kern)
        except Exception as error:
            ko_files.append(kern)
            print(f'Errors in {kern}: {error}')
            store_error_log(os.path.join(output_directory, 'errors.json'), {'kern': kern, 'error': str(error)})

    if remove_empty_directories:
        remove_empty_dirs(output_directory)

    print(f'----> OK files #{len(ok_files)}')
    print(f'----> KO files #{len(ko_files)}')
    print(ko_files)


if __name__ == '__main__':
    print(f'Usage: python -m kernpy --polish --input_directory /path/to/input --output_directory /path/to/output')
    sys.exit(1)

